# Copyright 2018 Rackspace, US Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import sys

from oslo_config import cfg
from oslo_db.sqlalchemy import enginefacade
import oslo_i18n as i18n
from oslo_log import log as logging

_translators = i18n.TranslatorFactory(domain='nlbaas2octavia')

# The primary translation function using the well-known name "_"
_ = _translators.primary

CONF = cfg.CONF

cli_opts = [
    cfg.BoolOpt('all', default=False,
                help='Migrate all load balancers'),
    cfg.StrOpt('lb_id',
               help='Load balancer ID to migrate'),
    cfg.StrOpt('project_id',
               help='Migrate all load balancers owned by this project'),
]

migration_opts = [
    cfg.BoolOpt('delete_after_migration', default=True,
                help='Delete the load balancer records from neutron-lbaas'
                     ' after migration'),
    cfg.BoolOpt('trial_run', default=False,
                help='Run without making changes.'),
    cfg.StrOpt('octavia_account_id', required=True,
               help='The keystone account ID Octavia is running under.'),
    cfg.StrOpt('neutron_db_connection',
               required=True,
               help='The neutron database connection string'),
    cfg.StrOpt('octavia_db_connection',
               required=True,
               help='The octavia database connection string'),
]

cfg.CONF.register_cli_opts(cli_opts)
cfg.CONF.register_opts(migration_opts, group='migration')


def cascade_delete_neutron_lb(n_session, lb_id):
    listeners = n_session.execute(
        "SELECT id FROM lbaas_listeners WHERE loadbalancer_id = :lb_id;",
        {'lb_id': lb_id})
    for listener in listeners:
        l7policies = n_session.execute(
            "SELECT id FROM lbaas_l7policies WHERE listener_id = :list_id;",
            {'list_id': listener[0]})
        for l7policy in l7policies:
            # Delete l7rules
            n_session.execute(
                "DELETE FROM lbaas_l7rules WHERE l7policy_id = :l7p_id;",
                {'l7p_id': l7policy[0]})
        # Delete l7policies
        n_session.execute(
            "DELETE FROM lbaas_l7policies WHERE listener_id = :list_id;",
            {'list_id': listener[0]})
        # Delete SNI records
        n_session.execute(
            "DELETE FROM lbaas_sni WHERE listener_id = :list_id;",
            {'list_id': listener[0]})

    # Delete the listeners
    n_session.execute(
        "DELETE FROM lbaas_listeners WHERE loadbalancer_id = :lb_id;",
        {'lb_id': lb_id})

    pools = n_session.execute(
        "SELECT id, healthmonitor_id FROM lbaas_pools "
        "WHERE loadbalancer_id = :lb_id;", {'lb_id': lb_id}).fetchall()
    for pool in pools:
        # Delete the members
        n_session.execute(
            "DELETE FROM lbaas_members WHERE pool_id = :pool_id;",
            {'pool_id': pool[0]})
        # Delete the session persistence records
        n_session.execute(
            "DELETE FROM lbaas_sessionpersistences WHERE pool_id = :pool_id;",
            {'pool_id': pool[0]})

        # Delete the pools
        n_session.execute(
            "DELETE FROM lbaas_pools WHERE id = :pool_id;",
            {'pool_id': pool[0]})

        # Delete the health monitor
        if pool[1]:
            result = n_session.execute("DELETE FROM lbaas_healthmonitors "
                                       "WHERE id = :id", {'id': pool[1]})
            if result.rowcount != 1:
                raise Exception(_('Failed to delete health monitor: '
                                '%s') % pool[1])
    # Delete the lb stats
    n_session.execute(
        "DELETE FROM lbaas_loadbalancer_statistics WHERE "
        "loadbalancer_id = :lb_id;", {'lb_id': lb_id})

    # Delete provider record
    n_session.execute(
        "DELETE FROM providerresourceassociations WHERE "
        "resource_id = :lb_id;", {'lb_id': lb_id})

    # Delete the load balanacer
    n_session.execute(
        "DELETE FROM lbaas_loadbalancers WHERE id = :lb_id;", {'lb_id': lb_id})


def process_health_monitor(LOG, n_session, o_session, project_id,
                           pool_id, hm_id):
    hm = n_session.execute(
        "SELECT type, delay, timeout, max_retries, http_method, url_path, "
        "expected_codes, admin_state_up, provisioning_status, name, "
        "max_retries_down FROM lbaas_healthmonitors WHERE id = :hm_id AND "
        "provisioning_status = 'ACTIVE';", {'hm_id': hm_id}).fetchone()
    LOG.debug('Migrating health manager: %s', hm_id)

    if hm is None:
        raise Exception(_('Health monitor %s has invalid '
                        'provisioning_status.'), hm_id)

    hm_op_status = 'ONLINE' if hm[7] else 'OFFLINE'

    result = o_session.execute(
        "INSERT INTO health_monitor (id, project_id, pool_id, type, delay, "
        "timeout, fall_threshold, rise_threshold, http_method, url_path, "
        "expected_codes, enabled, provisioning_status, name, created_at, "
        "updated_at, operating_status) VALUES (:id, :project_id, :pool_id, "
        ":type, :delay, :timeout, :fall_threshold, :rise_threshold, "
        ":http_method, :url_path, :expected_codes, :enabled, "
        ":provisioning_status, :name, :created_at, :updated_at, "
        ":operating_status);",
        {'id': hm_id, 'project_id': project_id, 'pool_id': pool_id,
         'type': hm[0], 'delay': hm[1], 'timeout': hm[2],
         'fall_threshold': hm[10], 'rise_threshold': hm[3],
         'http_method': hm[4], 'url_path': hm[5], 'expected_codes': hm[6],
         'enabled': hm[7], 'provisioning_status': hm[8], 'name': hm[9],
         'operating_status': hm_op_status,
         'created_at': datetime.datetime.utcnow(),
         'updated_at': datetime.datetime.utcnow()})
    if result.rowcount != 1:
        raise Exception(_('Unable to create health monitor in the Octavia '
                        'database.'))


def process_session_persistence(n_session, o_session, pool_id):
    # Setup session persistence if it is configured
    sp = n_session.execute(
        "SELECT type, cookie_name FROM lbaas_sessionpersistences "
        "WHERE pool_id = :pool_id;", {'pool_id': pool_id}).fetchone()
    if sp:
        result = o_session.execute(
            "INSERT INTO session_persistence (pool_id, type, cookie_name) "
            "VALUES (:pool_id, :type, :cookie_name);",
            {'pool_id': pool_id, 'type': sp[0], 'cookie_name': sp[1]})
        if result.rowcount != 1:
            raise Exception(_('Unable to create session persistence in the '
                            'Octavia database.'))


def process_members(LOG, n_session, o_session, project_id, pool_id):
    # Handle members
    members = n_session.execute(
        "SELECT id, subnet_id, address, protocol_port, weight, "
        "admin_state_up, provisioning_status, operating_status, name FROM "
        "lbaas_members WHERE pool_id = :pool_id;",
        {'pool_id': pool_id}).fetchall()
    for member in members:
        LOG.debug('Migrating member: %s', member[0])

        if member[6] == 'DELETED':
            continue
        elif member[6] != 'ACTIVE':
            raise Exception(_('Member is invalid state of %s.'), member[6])

        result = o_session.execute(
            "INSERT INTO member (id, pool_id, project_id, subnet_id, "
            "ip_address, protocol_port, weight, operating_status, enabled, "
            "created_at, updated_at, provisioning_status, name, backup) "
            "VALUES (:id, :pool_id, :project_id, :subnet_id, :ip_address, "
            ":protocol_port, :weight, :operating_status, :enabled, "
            ":created_at, :updated_at, :provisioning_status, :name, :backup);",
            {'id': member[0], 'pool_id': pool_id, 'project_id': project_id,
             'subnet_id': member[1], 'ip_address': member[2],
             'protocol_port': member[3], 'weight': member[4],
             'operating_status': member[7], 'enabled': member[5],
             'created_at': datetime.datetime.utcnow(),
             'updated_at': datetime.datetime.utcnow(),
             'provisioning_status': member[6], 'name': member[8],
             'backup': False})
    if result.rowcount != 1:
        raise Exception(_('Unable to create member in the Octavia database.'))


def process_SNI(n_session, o_session, listener_id):
    SNIs = n_session.execute(
        "SELECT tls_container_id, position FROM lbaas_sni WHERE "
        "listener_id = :listener_id;", {'listener_id': listener_id}).fetchall()
    for SNI in SNIs:
        result = o_session.execute(
            "INSERT INTO sni (listener_id, tls_container_id, position) VALUES "
            "(:listener_id, :tls_container_id, :position);",
            {'listener_id': listener_id, 'tls_container_id': SNI[0],
             'position': SNI[1]})
        if result.rowcount != 1:
            raise Exception(_('Unable to create SNI record in the Octavia '
                            'database.'))


def process_L7policies(LOG, n_session, o_session, listener_id, project_id):
    l7policies = n_session.execute(
        "SELECT id, name, description, listener_id, action, "
        "redirect_pool_id, redirect_url, position, "
        "provisioning_status, admin_state_up FROM "
        "lbaas_l7policies WHERE listener_id = :listener_id AND "
        "provisioning_status = 'ACTIVE';",
        {'listener_id': listener_id}).fetchall()
    for l7policy in l7policies:
        LOG.debug('Migrating L7 policy: %s', l7policy[0])

        if l7policy[8] == 'DELETED':
            continue
        elif l7policy[8] != 'ACTIVE':
            raise Exception(_('L7 policy is invalid state of %s.'),
                            l7policy[8])

        L7p_op_status = 'ONLINE' if l7policy[9] else 'OFFLINE'

        result = o_session.execute(
            "INSERT INTO l7policy (id, name, description, listener_id, "
            "action, redirect_pool_id, redirect_url, position, enabled, "
            "provisioning_status, created_at, updated_at, project_id, "
            "operating_status) VALUES (:id, :name, :description, "
            ":listener_id, :action, :redirect_pool_id, :redirect_url, "
            ":position, :enabled, :provisioning_status, :created_at, "
            ":updated_at, :project_id, :operating_status);",
            {'id': l7policy[0], 'name': l7policy[1],
             'description': l7policy[2], 'listener_id': listener_id,
             'action': l7policy[4], 'redirect_pool_id': l7policy[5],
             'redirect_url': l7policy[6], 'position': l7policy[7],
             'enabled': l7policy[9], 'provisioning_status': l7policy[8],
             'created_at': datetime.datetime.utcnow(),
             'updated_at': datetime.datetime.utcnow(),
             'project_id': project_id, 'operating_status': L7p_op_status})
        if result.rowcount != 1:
            raise Exception(_('Unable to create L7 policy in the Octavia '
                            'database.'))
        # Handle L7 rules
        l7rules = n_session.execute(
            "SELECT id, type, compare_type, invert, key, value, "
            "provisioning_status, admin_state_up FROM lbaas_l7rules WHERE "
            "l7policy_id = :l7policy_id AND provisioning_status = 'ACTIVE';",
            {'l7policy_id': l7policy[0]}).fetchall()
        for l7rule in l7rules:
            LOG.debug('Migrating L7 rule: %s', l7policy[0])

            if l7rule[6] == 'DELETED':
                continue
            elif l7rule[6] != 'ACTIVE':
                raise Exception(_('L7 rule is invalid state of %s.'),
                                l7rule[6])

            L7r_op_status = 'ONLINE' if l7rule[7] else 'OFFLINE'

            result = o_session.execute(
                "INSERT INTO l7rule (id, l7policy_id, type, compare_type, "
                "key, value, invert, provisioning_status, created_at, "
                "updated_at, project_id, enabled, operating_status) VALUES "
                "(:id, :l7policy_id, :type, :compare_type, :key, :value, "
                ":invert, :provisioning_status, :created_at, :updated_at, "
                ":project_id, :enabled, :operating_status);",
                {'id': l7rule[0], 'l7policy_id': l7policy[0],
                 'type': l7rule[1], 'compare_type': l7rule[2],
                 'key': l7rule[4], 'value': l7rule[5], 'invert': l7rule[3],
                 'provisioning_status': l7rule[6],
                 'created_at': datetime.datetime.utcnow(),
                 'updated_at': datetime.datetime.utcnow(),
                 'project_id': project_id, 'enabled': l7rule[7],
                 'operating_status': L7r_op_status})
            if result.rowcount != 1:
                raise Exception(_('Unable to create L7 policy in the Octavia '
                                'database.'))


def migrate_lb(LOG, n_session_maker, o_session_maker, lb_id):

    n_session = n_session_maker(autocommit=False)
    o_session = o_session_maker(autocommit=False)

    LOG.info('Migrating load balancer: %s', lb_id)
    try:
        # Lock the load balancer in neutron DB
        result = n_session.execute(
            "UPDATE lbaas_loadbalancers SET "
            "provisioning_status = 'PENDING_UPDATE' WHERE id = :id AND "
            "provisioning_status = 'ACTIVE';", {'id': lb_id})
        if result.rowcount != 1:
            raise Exception(_('Load balancer is not provisioning_status '
                            'ACTIVE'))

        # Get the load balancer record from neutron
        n_lb = n_session.execute(
            "SELECT b.provider_name, a.project_id, a.name, a.description, "
            "a.admin_state_up, a.operating_status, a.flavor_id, "
            "a.vip_port_id, a.vip_subnet_id, a.vip_address "
            "FROM lbaas_loadbalancers a JOIN providerresourceassociations b "
            "ON a.id = b.resource_id WHERE ID = :id;",
            {'id': lb_id}).fetchone()

        # Migrate the port and security groups to Octavia
        vip_port = n_session.execute(
            "SELECT a.device_owner, a.project_id, b.security_group_id "
            "FROM ports a JOIN securitygroupportbindings b ON "
            "a.id = b.port_id  where id = :id;",
            {'id': n_lb[7]}).fetchone()

        # neutron-lbaas does not support user VIP ports, so take
        # ownership of the port and security group
        if vip_port[0] == 'neutron:LOADBALANCERV2':
            result = n_session.execute(
                "UPDATE ports SET device_owner = 'Octavia', "
                "project_id = :proj_id WHERE "
                "id = :id;", {'id': n_lb[7],
                              'proj_id': CONF.migration.octavia_account_id})
            if result.rowcount != 1:
                raise Exception(_('Unable to update VIP port in the neutron '
                                'database.'))
            result = n_session.execute(
                "UPDATE securitygroups SET project_id = :proj_id WHERE "
                "id = :id;", {'proj_id': CONF.migration.octavia_account_id,
                              'id': vip_port[2]})
            if result.rowcount != 1:
                raise Exception(_('Unable to update VIP security group in the '
                                'neutron database.'))

        # Octavia driver load balancers are now done, next process the other
        # provider driver load balancers
        if n_lb[0] != 'octavia':
            # Create the load balancer
            result = o_session.execute(
                "INSERT INTO load_balancer (id, project_id, name, "
                "description, provisioning_status, operating_status, enabled, "
                "created_at, updated_at, provider) VALUES (:id, :project_id, "
                ":name, :description, :provisioning_status, "
                ":operating_status, :enabled, :created_at, :updated_at, "
                ":provider);",
                {'id': lb_id, 'project_id': n_lb[1], 'name': n_lb[2],
                 'description': n_lb[3], 'provisioning_status': 'ACTIVE',
                 'operating_status': n_lb[5], 'enabled': n_lb[4],
                 'created_at': datetime.datetime.utcnow(),
                 'updated_at': datetime.datetime.utcnow(),
                 'provider': n_lb[0]})
            if result.rowcount != 1:
                raise Exception(_('Unable to create load balancer in the '
                                'Octavia database.'))

            # Get the network ID for the VIP
            subnet = n_session.execute(
                "SELECT network_id FROM subnets WHERE id = :id;",
                {'id': n_lb[8]}).fetchone()

            # Create VIP record
            result = o_session.execute(
                "INSERT INTO vip (load_balancer_id, ip_address, port_id, "
                "subnet_id, network_id) VALUES (:lb_id, :ip_address, "
                ":port_id, :subnet_id, :network_id);",
                {'lb_id': lb_id, 'ip_address': n_lb[9], 'port_id': n_lb[7],
                 'subnet_id': n_lb[8], 'network_id': subnet[0]})
            if result.rowcount != 1:
                raise Exception(_('Unable to create VIP in the Octavia '
                                'database.'))

            # Create pools
            pools = n_session.execute(
                "SELECT id, name, description, protocol, lb_algorithm, "
                "healthmonitor_id, admin_state_up, provisioning_status, "
                "operating_status FROM lbaas_pools WHERE loadbalancer_id "
                " = :lb_id;",
                {'lb_id': lb_id}).fetchall()
            for pool in pools:
                LOG.debug('Migrating pool: %s', pool[0])

                if pool[7] == 'DELETED':
                    continue
                elif pool[7] != 'ACTIVE':
                    raise Exception(_('Pool is invalid state of %s.'), pool[7])

                result = o_session.execute(
                    "INSERT INTO pool (id, project_id, name, description, "
                    "protocol, lb_algorithm, operating_status, enabled, "
                    "load_balancer_id, created_at, updated_at, "
                    "provisioning_status) VALUES (:id, :project_id, :name, "
                    ":description, :protocol, :lb_algorithm, "
                    ":operating_status, :enabled, :load_balancer_id,"
                    ":created_at, :updated_at, :provisioning_status);",
                    {'id': pool[0], 'project_id': n_lb[1], 'name': pool[1],
                     'description': pool[2], 'protocol': pool[3],
                     'lb_algorithm': pool[4], 'operating_status': pool[8],
                     'enabled': pool[6], 'load_balancer_id': lb_id,
                     'created_at': datetime.datetime.utcnow(),
                     'updated_at': datetime.datetime.utcnow(),
                     'provisioning_status': pool[7]})
                if result.rowcount != 1:
                    raise Exception(_('Unable to create pool in the '
                                    'Octavia database.'))

                # Create health monitor if there is one
                if pool[5] is not None:
                    process_health_monitor(LOG, n_session, o_session,
                                           n_lb[1], pool[0], pool[5])

                # Handle the session persistence records
                process_session_persistence(n_session, o_session, pool[0])

                # Handle the pool memebers
                process_members(LOG, n_session, o_session, n_lb[1], pool[0])

            lb_stats = n_session.execute(
                "SELECT bytes_in, bytes_out, active_connections, "
                "total_connections FROM lbaas_loadbalancer_statistics WHERE "
                "loadbalancer_id = :lb_id;", {'lb_id': lb_id}).fetchone()
            listeners = n_session.execute(
                "SELECT id, name, description, protocol, protocol_port, "
                "connection_limit, default_pool_id, admin_state_up, "
                "provisioning_status, operating_status, "
                "default_tls_container_id FROM lbaas_listeners WHERE "
                "loadbalancer_id = :lb_id;", {'lb_id': lb_id}).fetchall()
            for listener in listeners:
                LOG.debug('Migrating listener: %s', listener[0])

                if listener[8] == 'DELETED':
                    continue
                elif listener[8] != 'ACTIVE':
                    raise Exception(_('Listener is invalid state of %s.'),
                                    listener[8])

                result = o_session.execute(
                    "INSERT INTO listener (id, project_id, name, description, "
                    "protocol, protocol_port, connection_limit, "
                    "load_balancer_id, tls_certificate_id, default_pool_id, "
                    "provisioning_status, operating_status, enabled, "
                    "created_at, updated_at) VALUES (:id, :project_id, :name, "
                    ":description, :protocol, :protocol_port, "
                    ":connection_limit, :load_balancer_id, "
                    ":tls_certificate_id, :default_pool_id, "
                    ":provisioning_status, :operating_status, :enabled, "
                    ":created_at, :updated_at);",
                    {'id': listener[0], 'project_id': n_lb[1],
                     'name': listener[1], 'description': listener[2],
                     'protocol': listener[3], 'protocol_port': listener[4],
                     'connection_limit': listener[5],
                     'load_balancer_id': lb_id,
                     'tls_certificate_id': listener[10],
                     'default_pool_id': listener[6],
                     'provisioning_status': listener[8],
                     'operating_status': listener[9], 'enabled': listener[7],
                     'created_at': datetime.datetime.utcnow(),
                     'updated_at': datetime.datetime.utcnow()})
                if result.rowcount != 1:
                    raise Exception(_('Unable to create listener in the '
                                    'Octavia database.'))

                # Convert load balancer stats to listener stats
                # This conversion may error on the low side due to
                # the division
                result = o_session.execute(
                    "INSERT INTO listener_statistics (listener_id, bytes_in, "
                    "bytes_out, active_connections, total_connections, "
                    "amphora_id, request_errors) VALUES (:listener_id, "
                    ":bytes_in, :bytes_out, :active_connections, "
                    ":total_connections, :amphora_id, :request_errors);",
                    {'listener_id': listener[0],
                     'bytes_in': int(lb_stats[0] / len(listeners)),
                     'bytes_out': int(lb_stats[1] / len(listeners)),
                     'active_connections': int(lb_stats[2] / len(listeners)),
                     'total_connections': int(lb_stats[3] / len(listeners)),
                     'amphora_id': listener[0], 'request_errors': 0})
                if result.rowcount != 1:
                    raise Exception(_('Unable to create listener statistics '
                                    'in the Octavia database.'))

                # Handle SNI certs
                process_SNI(n_session, o_session, listener[0])

                # Handle L7 policy records
                process_L7policies(LOG, n_session, o_session,
                                   listener[0], n_lb[1])

        # Delete the old neutron-lbaas records
        if (CONF.migration.delete_after_migration and not
                CONF.migration.trial_run):
            cascade_delete_neutron_lb(n_session, lb_id)

        if CONF.migration.trial_run:
            o_session.rollback()
            n_session.rollback()
            LOG.info('Simulated migration of load balancer %s successful.',
                     lb_id)
        else:
            o_session.commit()
            n_session.commit()
            LOG.info('Migration of load balancer %s successful.', lb_id)
        return 0
    except Exception as e:
        n_session.rollback()
        o_session.rollback()
        LOG.exception("Skipping load balancer %s due to: %s.", lb_id, str(e))
        return 1


def main():
    if len(sys.argv) == 1:
        print('Error: Config file must be specified.')
        print('nlbaas2octavia --config-file <filename>')
        return 1
    logging.register_options(cfg.CONF)
    cfg.CONF(args=sys.argv[1:],
             project='nlbaas2octavia',
             version='nlbaas2octavia 1.0')
    logging.set_defaults()
    logging.setup(cfg.CONF, 'nlbaas2octavia')
    LOG = logging.getLogger('nlbaas2octavia')
    CONF.log_opt_values(LOG, logging.DEBUG)

    if not CONF.all and not CONF.lb_id and not CONF.project_id:
        print('Error: One of --all, --lb_id, --project_id must be specified.')
        return 1

    if ((CONF.all and (CONF.lb_id or CONF.project_id)) or
            (CONF.lb_id and CONF.project_id)):
        print('Error: Only one of --all, --lb_id, --project_id allowed.')
        return 1

    neutron_context_manager = enginefacade.transaction_context()
    neutron_context_manager.configure(
        connection=CONF.migration.neutron_db_connection)
    n_session_maker = neutron_context_manager.writer.get_sessionmaker()

    octavia_context_manager = enginefacade.transaction_context()
    octavia_context_manager.configure(
        connection=CONF.migration.octavia_db_connection)
    o_session_maker = octavia_context_manager.writer.get_sessionmaker()

    LOG.info('Starting migration.')

    n_session = n_session_maker(autocommit=True)
    lb_id_list = []

    if CONF.lb_id:
        lb_id_list = [[CONF.lb_id]]
    elif CONF.project_id:
        lb_id_list = n_session.execute(
            "SELECT id FROM neutron.lbaas_loadbalancers WHERE "
            "project_id = :id AND provisioning_status = 'ACTIVE';",
            {'id': CONF.project_id}).fetchall()
    else:  # CONF.ALL
        lb_id_list = n_session.execute(
            "SELECT id FROM neutron.lbaas_loadbalancers WHERE "
            "provisioning_status = 'ACTIVE';").fetchall()

    failure_count = 0
    for lb in lb_id_list:
        failure_count += migrate_lb(LOG, n_session_maker,
                                    o_session_maker, lb[0])
    if failure_count:
        sys.exit(1)

if __name__ == "__main__":
    main()



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log.helpers import log_method_call as call_log

from neutron_lbaas.common import cert_manager
from neutron_lbaas.drivers import driver_base


class EdgeDriverBaseManager(object):
    @property
    def lbv2_driver(self):
        return self.driver.plugin.db._core_plugin.lbv2_driver


class EdgeLoadBalancerDriverV2(driver_base.LoadBalancerBaseDriver):
    @call_log
    def __init__(self, plugin):
        self.plugin = plugin
        super(EdgeLoadBalancerDriverV2, self).__init__(plugin)
        self.load_balancer = EdgeLoadBalancerManager(self)
        self.listener = EdgeListenerManager(self)
        self.pool = EdgePoolManager(self)
        self.member = EdgeMemberManager(self)
        self.health_monitor = EdgeHealthMonitorManager(self)
        self.l7policy = L7PolicyManager(self)
        self.l7rule = L7RuleManager(self)


class EdgeLoadBalancerManager(driver_base.BaseLoadBalancerManager,
                              EdgeDriverBaseManager):
    @call_log
    def create(self, context, lb):
        self.lbv2_driver.loadbalancer.create(context, lb)

    @call_log
    def update(self, context, old_lb, new_lb):
        self.lbv2_driver.loadbalancer.update(context, old_lb, new_lb)

    @call_log
    def delete(self, context, lb):
        self.lbv2_driver.loadbalancer.delete(context, lb)

    @call_log
    def refresh(self, context, lb):
        return self.lbv2_driver.loadbalancer.refresh(context, lb)

    @call_log
    def stats(self, context, lb):
        return self.lbv2_driver.loadbalancer.stats(context, lb)


class EdgeListenerManager(driver_base.BaseListenerManager,
                          EdgeDriverBaseManager):

    def _get_default_cert(self, listener):
        if listener.default_tls_container_id:
            cert_backend = cert_manager.get_backend()
            if cert_backend:
                return cert_backend.CertManager().get_cert(
                    project_id=listener.tenant_id,
                    cert_ref=listener.default_tls_container_id,
                    resource_ref=cert_backend.CertManager.get_service_url(
                        listener.loadbalancer_id)
                )

    @call_log
    def create(self, context, listener):
        self.lbv2_driver.listener.create(
            context, listener, certificate=self._get_default_cert(listener))

    @call_log
    def update(self, context, old_listener, new_listener):
        self.lbv2_driver.listener.update(
            context, old_listener, new_listener,
            certificate=self._get_default_cert(new_listener))

    @call_log
    def delete(self, context, listener):
        self.lbv2_driver.listener.delete(context, listener)


class EdgePoolManager(driver_base.BasePoolManager,
                      EdgeDriverBaseManager):
    @call_log
    def create(self, context, pool):
        self.lbv2_driver.pool.create(context, pool)

    @call_log
    def update(self, context, old_pool, new_pool):
        self.lbv2_driver.pool.update(context, old_pool, new_pool)

    @call_log
    def delete(self, context, pool):
        self.lbv2_driver.pool.delete(context, pool)


class EdgeMemberManager(driver_base.BaseMemberManager,
                        EdgeDriverBaseManager):
    @call_log
    def create(self, context, member):
        self.lbv2_driver.member.create(context, member)

    @call_log
    def update(self, context, old_member, new_member):
        self.lbv2_driver.member.update(context, old_member, new_member)

    @call_log
    def delete(self, context, member):
        self.lbv2_driver.member.delete(context, member)


class EdgeHealthMonitorManager(driver_base.BaseHealthMonitorManager,
                               EdgeDriverBaseManager):
    @call_log
    def create(self, context, hm):
        self.lbv2_driver.healthmonitor.create(context, hm)

    @call_log
    def update(self, context, old_hm, new_hm):
        self.lbv2_driver.healthmonitor.update(context, old_hm, new_hm)

    @call_log
    def delete(self, context, hm):
        self.lbv2_driver.healthmonitor.delete(context, hm)


class L7PolicyManager(driver_base.BaseL7PolicyManager,
                      EdgeDriverBaseManager):
    @call_log
    def create(self, context, pol):
        self.lbv2_driver.l7policy.create(context, pol)

    @call_log
    def update(self, context, old_pol, new_pol):
        self.lbv2_driver.l7policy.update(context, old_pol, new_pol)

    @call_log
    def delete(self, context, pol):
        self.lbv2_driver.l7policy.delete(context, pol)


class L7RuleManager(driver_base.BaseL7RuleManager,
                    EdgeDriverBaseManager):
    @call_log
    def create(self, context, rule):
        self.lbv2_driver.l7rule.create(context, rule)

    @call_log
    def update(self, context, old_rule, new_rule):
        self.lbv2_driver.l7rule.update(context, old_rule, new_rule)

    @call_log
    def delete(self, context, rule):
        self.lbv2_driver.l7rule.delete(context, rule)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

import mock
import testscenarios

from neutron_lbaas.services.loadbalancer import data_models
from neutron_lbaas.tests import base
from neutron_lbaas.tests import tools

load_tests = testscenarios.load_tests_apply_scenarios


class TestBaseDataModel(base.BaseTestCase):

    def _get_fake_model_cls(self, fields_):
        class FakeModel(data_models.BaseDataModel):
            fields = fields_

            def __init__(self, **kwargs):
                for k, v in kwargs.items():
                    setattr(self, k, v)

        return FakeModel

    def test_from_dict(self):

        fields_ = ['field1', 'field2']
        dict_ = {field: tools.get_random_string()
                 for field in fields_}

        model_cls = self._get_fake_model_cls(fields_)
        model = model_cls.from_dict(dict_)

        for field in fields_:
            self.assertEqual(dict_[field], getattr(model, field))

    def test_from_dict_filters_by_fields(self):

        fields_ = ['field1', 'field2']
        dict_ = {field: tools.get_random_string()
                 for field in fields_}
        dict_['foo'] = 'bar'

        model_cls = self._get_fake_model_cls(fields_)
        model = model_cls.from_dict(dict_)
        self.assertFalse(hasattr(model, 'foo'))


def _get_models():
    models = []
    for name, obj in inspect.getmembers(data_models):
        if inspect.isclass(obj):
            if issubclass(obj, data_models.BaseDataModel):
                if type(obj) != data_models.BaseDataModel:
                    models.append(obj)
    return models


class TestModels(base.BaseTestCase):

    scenarios = [
        (model.__name__, {'model': model})
        for model in _get_models()
    ]

    @staticmethod
    def _get_iterable_mock(*args, **kwargs):
        m = mock.create_autospec(dict, spec_set=True)

        def _get_empty_iterator(*args, **kwargs):
            return iter([])

        m.__iter__ = _get_empty_iterator
        m.pop = _get_empty_iterator
        return m

    def test_from_dict_filters_by_fields(self):

        dict_ = {field: self._get_iterable_mock()
                 for field in self.model.fields}
        dict_['foo'] = 'bar'

        model = self.model.from_dict(dict_)
        self.assertFalse(hasattr(model, 'foo'))



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Add flavor id

Revision ID: 3426acbc12de
Revises: 4a408dd491c2
Create Date: 2015-12-02 15:24:35.775474

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3426acbc12de'
down_revision = '4a408dd491c2'


def upgrade():
    op.add_column('lbaas_loadbalancers',
                  sa.Column(u'flavor_id', sa.String(36), nullable=True))
    op.create_foreign_key(u'fk_lbaas_loadbalancers_flavors_id',
                          u'lbaas_loadbalancers',
                          u'flavors',
                          [u'flavor_id'],
                          [u'id'])



# Copyright 2013 New Dream Network, LLC (DreamHost)
# Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

import mock
from neutron_lib import constants

from neutron_lbaas.agent import agent_manager as manager
from neutron_lbaas.services.loadbalancer import constants as lb_const
from neutron_lbaas.services.loadbalancer import data_models
from neutron_lbaas.tests import base


class TestManager(base.BaseTestCase):
    def setUp(self):
        super(TestManager, self).setUp()

        mock_conf = mock.Mock()
        mock_conf.device_driver = ['devdriver']

        self.mock_importer = mock.patch.object(manager, 'importutils').start()

        rpc_mock_cls = mock.patch(
            'neutron_lbaas.agent.agent_api.LbaasAgentApi'
        ).start()

        # disable setting up periodic state reporting
        mock_conf.AGENT.report_interval = 0

        self.mgr = manager.LbaasAgentManager(mock_conf)
        self.rpc_mock = rpc_mock_cls.return_value
        self.log = mock.patch.object(manager, 'LOG').start()
        self.driver_mock = mock.Mock()
        self.mgr.device_drivers = {'devdriver': self.driver_mock}
        instance_mapping = [('1', 'devdriver'), ('2', 'devdriver')]
        self.mgr.instance_mapping = collections.OrderedDict(instance_mapping)
        self.mgr.needs_resync = False
        self.update_statuses_patcher = mock.patch.object(
            self.mgr, '_update_statuses')
        self.update_statuses = self.update_statuses_patcher.start()

    def test_initialize_service_hook(self):
        with mock.patch.object(self.mgr, 'sync_state') as sync:
            self.mgr.initialize_service_hook(mock.Mock())
            sync.assert_called_once_with()

    def test_periodic_resync_needs_sync(self):
        with mock.patch.object(self.mgr, 'sync_state') as sync:
            self.mgr.needs_resync = True
            self.mgr.periodic_resync(mock.Mock())
            sync.assert_called_once_with()

    def test_periodic_resync_no_sync(self):
        with mock.patch.object(self.mgr, 'sync_state') as sync:
            self.mgr.needs_resync = False
            self.mgr.periodic_resync(mock.Mock())
            self.assertFalse(sync.called)

    def test_collect_stats(self):
        self.mgr.collect_stats(mock.Mock())
        self.rpc_mock.update_loadbalancer_stats.assert_has_calls([
            mock.call('1', mock.ANY),
            mock.call('2', mock.ANY)
        ], any_order=True)

    def test_collect_stats_exception(self):
        self.driver_mock.loadbalancer.get_stats.side_effect = Exception

        self.mgr.collect_stats(mock.Mock())

        self.assertFalse(self.rpc_mock.called)
        self.assertTrue(self.mgr.needs_resync)
        self.assertTrue(self.log.exception.called)

    def _sync_state_helper(self, ready, reloaded, destroyed):
        with mock.patch.object(self.mgr, '_reload_loadbalancer') as reload, \
                mock.patch.object(self.mgr, '_destroy_loadbalancer') as \
                destroy:
            self.rpc_mock.get_ready_devices.return_value = ready

            self.mgr.sync_state()

            self.assertEqual(len(reloaded), len(reload.mock_calls))
            self.assertEqual(len(destroyed), len(destroy.mock_calls))

            reload.assert_has_calls([mock.call(i) for i in reloaded],
                                    any_order=True)
            destroy.assert_has_calls([mock.call(i) for i in destroyed],
                                     any_order=True)
            self.assertFalse(self.mgr.needs_resync)

    def test_sync_state_all_known(self):
        self._sync_state_helper(['1', '2'], ['1', '2'], [])

    def test_sync_state_all_unknown(self):
        self.mgr.instance_mapping = {}
        self._sync_state_helper(['1', '2'], ['1', '2'], [])

    def test_sync_state_destroy_all(self):
        self._sync_state_helper([], [], ['1', '2'])

    def test_sync_state_both(self):
        self.mgr.instance_mapping = {'1': 'devdriver'}
        self._sync_state_helper(['2'], ['2'], ['1'])

    def test_sync_state_exception(self):
        self.rpc_mock.get_ready_devices.side_effect = Exception

        self.mgr.sync_state()

        self.assertTrue(self.log.exception.called)
        self.assertTrue(self.mgr.needs_resync)

    def test_reload_loadbalancer(self):
        lb = data_models.LoadBalancer(id='1').to_dict()
        lb['provider'] = {'device_driver': 'devdriver'}
        self.rpc_mock.get_loadbalancer.return_value = lb
        lb_id = 'new_id'
        self.assertNotIn(lb_id, self.mgr.instance_mapping)

        self.mgr._reload_loadbalancer(lb_id)

        calls = self.driver_mock.deploy_instance.call_args_list
        self.assertEqual(1, len(calls))
        called_lb = calls[0][0][0]
        self.assertEqual(lb['id'], called_lb.id)
        self.assertIn(lb['id'], self.mgr.instance_mapping)
        self.rpc_mock.loadbalancer_deployed.assert_called_once_with(lb_id)

    def test_reload_loadbalancer_driver_not_found(self):
        lb = data_models.LoadBalancer(id='1').to_dict()
        lb['provider'] = {'device_driver': 'unknowndriver'}
        self.rpc_mock.get_loadbalancer.return_value = lb
        lb_id = 'new_id'
        self.assertNotIn(lb_id, self.mgr.instance_mapping)

        self.mgr._reload_loadbalancer(lb_id)

        self.assertTrue(self.log.error.called)
        self.assertFalse(self.driver_mock.deploy_instance.called)
        self.assertNotIn(lb_id, self.mgr.instance_mapping)
        self.assertFalse(self.rpc_mock.loadbalancer_deployed.called)

    def test_reload_loadbalancer_exception_on_driver(self):
        lb = data_models.LoadBalancer(id='3').to_dict()
        lb['provider'] = {'device_driver': 'devdriver'}
        self.rpc_mock.get_loadbalancer.return_value = lb
        self.driver_mock.deploy_instance.side_effect = Exception
        lb_id = 'new_id'
        self.assertNotIn(lb_id, self.mgr.instance_mapping)

        self.mgr._reload_loadbalancer(lb_id)

        calls = self.driver_mock.deploy_instance.call_args_list
        self.assertEqual(1, len(calls))
        called_lb = calls[0][0][0]
        self.assertEqual(lb['id'], called_lb.id)
        self.assertNotIn(lb['id'], self.mgr.instance_mapping)
        self.assertFalse(self.rpc_mock.loadbalancer_deployed.called)
        self.assertTrue(self.log.exception.called)
        self.assertTrue(self.mgr.needs_resync)

    def test_destroy_loadbalancer(self):
        lb_id = '1'
        self.assertIn(lb_id, self.mgr.instance_mapping)

        self.mgr._destroy_loadbalancer(lb_id)

        self.driver_mock.undeploy_instance.assert_called_once_with(
            lb_id, delete_namespace=True)
        self.assertNotIn(lb_id, self.mgr.instance_mapping)
        self.rpc_mock.loadbalancer_destroyed.assert_called_once_with(lb_id)
        self.assertFalse(self.mgr.needs_resync)

    def test_destroy_loadbalancer_exception_on_driver(self):
        lb_id = '1'
        self.assertIn(lb_id, self.mgr.instance_mapping)
        self.driver_mock.undeploy_instance.side_effect = Exception

        self.mgr._destroy_loadbalancer(lb_id)

        self.driver_mock.undeploy_instance.assert_called_once_with(
            lb_id, delete_namespace=True)
        self.assertIn(lb_id, self.mgr.instance_mapping)
        self.assertFalse(self.rpc_mock.loadbalancer_destroyed.called)
        self.assertTrue(self.log.exception.called)
        self.assertTrue(self.mgr.needs_resync)

    def test_get_driver_unknown_device(self):
        self.assertRaises(manager.DeviceNotFoundOnAgent,
                          self.mgr._get_driver, 'unknown')

    def test_remove_orphans(self):
        self.mgr.remove_orphans()
        self.driver_mock.remove_orphans.assert_called_once_with(['1', '2'])

    def test_agent_disabled(self):
        payload = {'admin_state_up': False}
        self.mgr.agent_updated(mock.Mock(), payload)
        self.driver_mock.undeploy_instance.assert_has_calls(
            [mock.call('1', delete_namespace=True),
             mock.call('2', delete_namespace=True)],
            any_order=True
        )

    def test_update_statuses_loadbalancer(self):
        self.update_statuses_patcher.stop()
        lb = data_models.LoadBalancer(id='1')
        self.mgr._update_statuses(lb)
        self.rpc_mock.update_status.assert_called_once_with(
            'loadbalancer', lb.id, provisioning_status=constants.ACTIVE,
            operating_status=lb_const.ONLINE)

        self.rpc_mock.update_status.reset_mock()
        self.mgr._update_statuses(lb, error=True)
        self.rpc_mock.update_status.assert_called_once_with(
            'loadbalancer', lb.id, provisioning_status=constants.ERROR,
            operating_status=lb_const.OFFLINE)

    def test_update_statuses_listener(self):
        self.update_statuses_patcher.stop()
        listener = data_models.Listener(id='1')
        lb = data_models.LoadBalancer(id='1', listeners=[listener])
        listener.loadbalancer = lb
        self.mgr._update_statuses(listener)
        self.assertEqual(2, self.rpc_mock.update_status.call_count)
        calls = [mock.call('listener', listener.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=lb_const.ONLINE),
                 mock.call('loadbalancer', lb.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=None)]
        self.rpc_mock.update_status.assert_has_calls(calls)

        self.rpc_mock.update_status.reset_mock()
        self.mgr._update_statuses(listener, error=True)
        self.assertEqual(2, self.rpc_mock.update_status.call_count)
        calls = [mock.call('listener', listener.id,
                           provisioning_status=constants.ERROR,
                           operating_status=lb_const.OFFLINE),
                 mock.call('loadbalancer', lb.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=None)]
        self.rpc_mock.update_status.assert_has_calls(calls)

    def test_update_statuses_pool(self):
        self.update_statuses_patcher.stop()
        pool = data_models.Pool(id='1')
        listener = data_models.Listener(id='1', default_pool=pool)
        lb = data_models.LoadBalancer(id='1', listeners=[listener])
        listener.loadbalancer = lb
        pool.loadbalancer = lb
        self.mgr._update_statuses(pool)
        self.assertEqual(2, self.rpc_mock.update_status.call_count)
        calls = [mock.call('pool', pool.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=lb_const.ONLINE),
                 mock.call('loadbalancer', lb.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=None)]
        self.rpc_mock.update_status.assert_has_calls(calls)

        self.rpc_mock.update_status.reset_mock()
        self.mgr._update_statuses(pool, error=True)
        self.assertEqual(2, self.rpc_mock.update_status.call_count)
        calls = [mock.call('pool', pool.id,
                           provisioning_status=constants.ERROR,
                           operating_status=lb_const.OFFLINE),
                 mock.call('loadbalancer', lb.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=None)]
        self.rpc_mock.update_status.assert_has_calls(calls)

    def test_update_statuses_member(self):
        self.update_statuses_patcher.stop()
        member = data_models.Member(id='1')
        pool = data_models.Pool(id='1', members=[member])
        member.pool = pool
        listener = data_models.Listener(id='1', default_pool=pool)
        lb = data_models.LoadBalancer(id='1', listeners=[listener])
        listener.loadbalancer = lb
        pool.loadbalancer = lb
        self.mgr._update_statuses(member)
        self.assertEqual(2, self.rpc_mock.update_status.call_count)
        calls = [mock.call('member', member.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=lb_const.ONLINE),
                 mock.call('loadbalancer', lb.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=None)]
        self.rpc_mock.update_status.assert_has_calls(calls)

        self.rpc_mock.update_status.reset_mock()
        self.mgr._update_statuses(member, error=True)
        self.assertEqual(2, self.rpc_mock.update_status.call_count)
        calls = [mock.call('member', member.id,
                           provisioning_status=constants.ERROR,
                           operating_status=lb_const.OFFLINE),
                 mock.call('loadbalancer', lb.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=None)]
        self.rpc_mock.update_status.assert_has_calls(calls)

    def test_update_statuses_healthmonitor(self):
        self.update_statuses_patcher.stop()
        hm = data_models.HealthMonitor(id='1')
        pool = data_models.Pool(id='1', healthmonitor=hm)
        hm.pool = pool
        listener = data_models.Listener(id='1', default_pool=pool)
        lb = data_models.LoadBalancer(id='1', listeners=[listener])
        listener.loadbalancer = lb
        pool.loadbalancer = lb
        self.mgr._update_statuses(hm)
        self.assertEqual(2, self.rpc_mock.update_status.call_count)
        calls = [mock.call('healthmonitor', hm.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=None),
                 mock.call('loadbalancer', lb.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=None)]
        self.rpc_mock.update_status.assert_has_calls(calls)

        self.rpc_mock.update_status.reset_mock()
        self.mgr._update_statuses(hm, error=True)
        self.assertEqual(2, self.rpc_mock.update_status.call_count)
        calls = [mock.call('healthmonitor', hm.id,
                           provisioning_status=constants.ERROR,
                           operating_status=None),
                 mock.call('loadbalancer', lb.id,
                           provisioning_status=constants.ACTIVE,
                           operating_status=None)]
        self.rpc_mock.update_status.assert_has_calls(calls)

    @mock.patch.object(data_models.LoadBalancer, 'from_dict')
    def test_create_loadbalancer(self, mlb):
        loadbalancer = data_models.LoadBalancer(id='1')

        self.assertIn(loadbalancer.id, self.mgr.instance_mapping)
        mlb.return_value = loadbalancer
        self.mgr.create_loadbalancer(mock.Mock(), loadbalancer.to_dict(),
                                     'devdriver')
        self.driver_mock.loadbalancer.create.assert_called_once_with(
            loadbalancer)
        self.update_statuses.assert_called_once_with(loadbalancer)

    @mock.patch.object(data_models.LoadBalancer, 'from_dict')
    def test_create_loadbalancer_failed(self, mlb):
        loadbalancer = data_models.LoadBalancer(id='1')

        self.assertIn(loadbalancer.id, self.mgr.instance_mapping)
        self.driver_mock.loadbalancer.create.side_effect = Exception
        mlb.return_value = loadbalancer
        self.mgr.create_loadbalancer(mock.Mock(), loadbalancer.to_dict(),
                                     'devdriver')
        self.driver_mock.loadbalancer.create.assert_called_once_with(
            loadbalancer)
        self.update_statuses.assert_called_once_with(loadbalancer, error=True)

    @mock.patch.object(data_models.LoadBalancer, 'from_dict')
    def test_update_loadbalancer(self, mlb):

        loadbalancer = data_models.LoadBalancer(id='1',
                                                vip_address='10.0.0.1')
        old_loadbalancer = data_models.LoadBalancer(id='1',
                                                    vip_address='10.0.0.2')

        mlb.side_effect = [loadbalancer, old_loadbalancer]
        self.mgr.update_loadbalancer(mock.Mock(), old_loadbalancer.to_dict(),
                                     loadbalancer.to_dict())
        self.driver_mock.loadbalancer.update.assert_called_once_with(
            old_loadbalancer, loadbalancer)
        self.update_statuses.assert_called_once_with(loadbalancer)

    @mock.patch.object(data_models.LoadBalancer, 'from_dict')
    def test_update_loadbalancer_failed(self, mlb):
        loadbalancer = data_models.LoadBalancer(id='1',
                                                vip_address='10.0.0.1')
        old_loadbalancer = data_models.LoadBalancer(id='1',
                                                    vip_address='10.0.0.2')

        mlb.side_effect = [loadbalancer, old_loadbalancer]
        self.driver_mock.loadbalancer.update.side_effect = Exception
        self.mgr.update_loadbalancer(mock.Mock(), old_loadbalancer,
                                     loadbalancer)
        self.driver_mock.loadbalancer.update.assert_called_once_with(
            old_loadbalancer, loadbalancer)
        self.update_statuses.assert_called_once_with(loadbalancer, error=True)

    @mock.patch.object(data_models.LoadBalancer, 'from_dict')
    def test_delete_loadbalancer(self, mlb):
        loadbalancer = data_models.LoadBalancer(id='1')
        mlb.return_value = loadbalancer
        self.assertIn(loadbalancer.id, self.mgr.instance_mapping)
        self.mgr.delete_loadbalancer(mock.Mock(), loadbalancer.to_dict())
        self.driver_mock.loadbalancer.delete.assert_called_once_with(
            loadbalancer)
        self.assertNotIn(loadbalancer.id, self.mgr.instance_mapping)

    @mock.patch.object(data_models.Listener, 'from_dict')
    def test_create_listener(self, mlistener):
        loadbalancer = data_models.LoadBalancer(id='1')
        listener = data_models.Listener(id=1, loadbalancer_id='1',
                                        loadbalancer=loadbalancer)

        self.assertIn(loadbalancer.id, self.mgr.instance_mapping)
        mlistener.return_value = listener
        self.mgr.create_listener(mock.Mock(), listener.to_dict())
        self.driver_mock.listener.create.assert_called_once_with(listener)
        self.update_statuses.assert_called_once_with(listener)

    @mock.patch.object(data_models.Listener, 'from_dict')
    def test_create_listener_failed(self, mlistener):
        loadbalancer = data_models.LoadBalancer(id='1')
        listener = data_models.Listener(id=1, loadbalancer_id='1',
                                        loadbalancer=loadbalancer)

        self.assertIn(loadbalancer.id, self.mgr.instance_mapping)
        self.driver_mock.listener.create.side_effect = Exception
        mlistener.return_value = listener
        self.mgr.create_listener(mock.Mock(), listener.to_dict())
        self.driver_mock.listener.create.assert_called_once_with(listener)
        self.update_statuses.assert_called_once_with(listener, error=True)

    @mock.patch.object(data_models.Listener, 'from_dict')
    def test_update_listener(self, mlistener):
        loadbalancer = data_models.LoadBalancer(id='1')
        old_listener = data_models.Listener(id=1, loadbalancer_id='1',
                                            loadbalancer=loadbalancer,
                                            protocol_port=80)
        listener = data_models.Listener(id=1, loadbalancer_id='1',
                                        loadbalancer=loadbalancer,
                                        protocol_port=81)

        mlistener.side_effect = [listener, old_listener]
        self.mgr.update_listener(mock.Mock(), old_listener.to_dict(),
                                 listener.to_dict())
        self.driver_mock.listener.update.assert_called_once_with(
            old_listener, listener)
        self.update_statuses.assert_called_once_with(listener)

    @mock.patch.object(data_models.Listener, 'from_dict')
    def test_update_listener_failed(self, mlistener):
        loadbalancer = data_models.LoadBalancer(id='1')
        old_listener = data_models.Listener(id=1, loadbalancer_id='1',
                                            loadbalancer=loadbalancer,
                                            protocol_port=80)
        listener = data_models.Listener(id=1, loadbalancer_id='1',
                                        loadbalancer=loadbalancer,
                                        protocol_port=81)
        mlistener.side_effect = [listener, old_listener]
        self.driver_mock.listener.update.side_effect = Exception
        self.mgr.update_listener(mock.Mock(), old_listener, listener)
        self.driver_mock.listener.update.assert_called_once_with(old_listener,
                                                                 listener)
        self.update_statuses.assert_called_once_with(listener, error=True)

    @mock.patch.object(data_models.Listener, 'from_dict')
    def test_delete_listener(self, mlistener):
        loadbalancer = data_models.LoadBalancer(id='1')
        listener = data_models.Listener(id=1, loadbalancer_id='1',
                                        loadbalancer=loadbalancer,
                                        protocol_port=80)
        mlistener.return_value = listener
        self.mgr.delete_listener(mock.Mock(), listener.to_dict())
        self.driver_mock.listener.delete.assert_called_once_with(listener)

    @mock.patch.object(data_models.Pool, 'from_dict')
    def test_create_pool(self, mpool):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer)

        mpool.return_value = pool
        self.mgr.create_pool(mock.Mock(), pool.to_dict())
        self.driver_mock.pool.create.assert_called_once_with(pool)
        self.update_statuses.assert_called_once_with(pool)

    @mock.patch.object(data_models.Pool, 'from_dict')
    def test_create_pool_failed(self, mpool):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer)

        mpool.return_value = pool
        self.driver_mock.pool.create.side_effect = Exception
        self.mgr.create_pool(mock.Mock(), pool)
        self.driver_mock.pool.create.assert_called_once_with(pool)
        self.update_statuses.assert_called_once_with(pool, error=True)

    @mock.patch.object(data_models.Pool, 'from_dict')
    def test_update_pool(self, mpool):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        old_pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                    protocol='HTTP')
        mpool.side_effect = [pool, old_pool]
        self.mgr.update_pool(mock.Mock(), old_pool.to_dict(), pool.to_dict())
        self.driver_mock.pool.update.assert_called_once_with(old_pool, pool)
        self.update_statuses.assert_called_once_with(pool)

    @mock.patch.object(data_models.Pool, 'from_dict')
    def test_update_pool_failed(self, mpool):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        old_pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                    protocol='HTTP')
        mpool.side_effect = [pool, old_pool]
        self.driver_mock.pool.update.side_effect = Exception
        self.mgr.update_pool(mock.Mock(), old_pool.to_dict(), pool.to_dict())
        self.driver_mock.pool.update.assert_called_once_with(old_pool, pool)
        self.update_statuses.assert_called_once_with(pool, error=True)

    @mock.patch.object(data_models.Pool, 'from_dict')
    def test_delete_pool(self, mpool):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        mpool.return_value = pool
        self.mgr.delete_pool(mock.Mock(), pool.to_dict())
        self.driver_mock.pool.delete.assert_called_once_with(pool)

    @mock.patch.object(data_models.Member, 'from_dict')
    def test_create_member(self, mmember):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        member = data_models.Member(id='1', pool=pool)
        mmember.return_value = member
        self.mgr.create_member(mock.Mock(), member.to_dict())
        self.driver_mock.member.create.assert_called_once_with(member)
        self.update_statuses.assert_called_once_with(member)

    @mock.patch.object(data_models.Member, 'from_dict')
    def test_create_member_failed(self, mmember):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        member = data_models.Member(id='1', pool=pool)
        mmember.return_value = member
        self.driver_mock.member.create.side_effect = Exception
        self.mgr.create_member(mock.Mock(), member.to_dict())
        self.driver_mock.member.create.assert_called_once_with(member)
        self.update_statuses.assert_called_once_with(member, error=True)

    @mock.patch.object(data_models.Member, 'from_dict')
    def test_update_member(self, mmember):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        member = data_models.Member(id='1', pool=pool, weight=1)
        old_member = data_models.Member(id='1', pool=pool, weight=2)
        mmember.side_effect = [member, old_member]
        self.mgr.update_member(mock.Mock(), old_member.to_dict(),
                               member.to_dict())
        self.driver_mock.member.update.assert_called_once_with(old_member,
                                                               member)
        self.update_statuses.assert_called_once_with(member)

    @mock.patch.object(data_models.Member, 'from_dict')
    def test_update_member_failed(self, mmember):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        member = data_models.Member(id='1', pool=pool, weight=1)
        old_member = data_models.Member(id='1', pool=pool, weight=2)
        mmember.side_effect = [member, old_member]
        self.driver_mock.member.update.side_effect = Exception
        self.mgr.update_member(mock.Mock(), old_member.to_dict(),
                               member.to_dict())
        self.driver_mock.member.update.assert_called_once_with(old_member,
                                                               member)
        self.update_statuses.assert_called_once_with(member, error=True)

    @mock.patch.object(data_models.Member, 'from_dict')
    def test_delete_member(self, mmember):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        member = data_models.Member(id='1', pool=pool, weight=1)
        mmember.return_value = member
        self.mgr.delete_member(mock.Mock(), member.to_dict())
        self.driver_mock.member.delete.assert_called_once_with(member)

    @mock.patch.object(data_models.HealthMonitor, 'from_dict')
    def test_create_monitor(self, mmonitor):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        monitor = data_models.HealthMonitor(id='1', pool=pool)
        mmonitor.return_value = monitor
        self.mgr.create_healthmonitor(mock.Mock(), monitor.to_dict())
        self.driver_mock.healthmonitor.create.assert_called_once_with(
            monitor)
        self.update_statuses.assert_called_once_with(monitor)

    @mock.patch.object(data_models.HealthMonitor, 'from_dict')
    def test_create_monitor_failed(self, mmonitor):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        monitor = data_models.HealthMonitor(id='1', pool=pool)
        mmonitor.return_value = monitor
        self.driver_mock.healthmonitor.create.side_effect = Exception
        self.mgr.create_healthmonitor(mock.Mock(), monitor.to_dict())
        self.driver_mock.healthmonitor.create.assert_called_once_with(monitor)
        self.update_statuses.assert_called_once_with(monitor, error=True)

    @mock.patch.object(data_models.HealthMonitor, 'from_dict')
    def test_update_monitor(self, mmonitor):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        monitor = data_models.HealthMonitor(id='1', pool=pool, delay=1)
        old_monitor = data_models.HealthMonitor(id='1', pool=pool, delay=2)
        mmonitor.side_effect = [monitor, old_monitor]
        self.mgr.update_healthmonitor(mock.Mock(), old_monitor.to_dict(),
                                      monitor.to_dict())
        self.driver_mock.healthmonitor.update.assert_called_once_with(
            old_monitor, monitor)
        self.update_statuses.assert_called_once_with(monitor)

    @mock.patch.object(data_models.HealthMonitor, 'from_dict')
    def test_update_monitor_failed(self, mmonitor):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        monitor = data_models.HealthMonitor(id='1', pool=pool, delay=1)
        old_monitor = data_models.HealthMonitor(id='1', pool=pool, delay=2)
        mmonitor.side_effect = [monitor, old_monitor]
        self.driver_mock.healthmonitor.update.side_effect = Exception
        self.mgr.update_healthmonitor(mock.Mock(), monitor.to_dict(),
                                      monitor.to_dict())
        self.driver_mock.healthmonitor.update.assert_called_once_with(
            old_monitor, monitor)
        self.update_statuses.assert_called_once_with(monitor, error=True)

    @mock.patch.object(data_models.HealthMonitor, 'from_dict')
    def test_delete_monitor(self, mmonitor):
        loadbalancer = data_models.LoadBalancer(id='1')
        pool = data_models.Pool(id='1', loadbalancer=loadbalancer,
                                protocol='HTTPS')
        monitor = data_models.HealthMonitor(id='1', pool=pool)
        mmonitor.return_value = monitor
        self.mgr.delete_healthmonitor(mock.Mock(), monitor.to_dict())
        self.driver_mock.healthmonitor.delete.assert_called_once_with(
            monitor)



# Copyright 2015, 2016 Rackspace US Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as ex

from neutron_lbaas.tests.tempest.v2.api import base


class ListenersTestJSON(base.BaseTestCase):

    """
    Tests the following operations in the Neutron-LBaaS API using the
    REST client for Listeners:

        list listeners
        create listener
        get listener
        update listener
        delete listener
    """

    @classmethod
    def resource_setup(cls):
        super(ListenersTestJSON, cls).resource_setup()
        if not utils.is_extension_enabled('lbaasv2', 'network'):
            msg = "lbaas extension not enabled."
            raise cls.skipException(msg)
        network_name = data_utils.rand_name('network-')
        cls.network = cls.create_network(network_name)
        cls.subnet = cls.create_subnet(cls.network)
        cls.create_lb_kwargs = {'tenant_id': cls.subnet['tenant_id'],
                                'vip_subnet_id': cls.subnet['id']}
        cls.load_balancer = cls._create_active_load_balancer(
            **cls.create_lb_kwargs)
        cls.port = 80
        cls.load_balancer_id = cls.load_balancer['id']
        cls.create_listener_kwargs = {'loadbalancer_id': cls.load_balancer_id,
                                      'protocol': cls.listener_protocol,
                                      'protocol_port': cls.port}
        cls.listener = cls._create_listener(**cls.create_listener_kwargs)
        cls.listener_id = cls.listener['id']

    @decorators.attr(type='smoke')
    def test_get_listener(self):
        """Test get listener"""
        listener = self.listeners_client.get_listener(
            self.listener_id)
        self._test_provisioning_status_if_exists(self.listener, listener)
        self.assertEqual(self.listener, listener)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    def _prep_list_comparison(self, single, obj_list):
        for obj in obj_list:
            if not single.get('updated_at') and obj.get('updated_at'):
                obj['updated_at'] = None
            self._test_provisioning_status_if_exists(single, obj)

    def test_list_listeners(self):
        """Test get listeners with one listener"""
        listeners = self.listeners_client.list_listeners()
        self.assertEqual(len(listeners), 1)
        self._prep_list_comparison(self.listener, listeners)
        self.assertIn(self.listener, listeners)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='smoke')
    def test_list_listeners_two(self):
        """Test get listeners with two listeners"""
        create_new_listener_kwargs = self.create_listener_kwargs
        create_new_listener_kwargs['protocol_port'] = 8080
        new_listener = self._create_listener(
            **create_new_listener_kwargs)
        new_listener_id = new_listener['id']
        self.addCleanup(self._delete_listener, new_listener_id)
        self._check_status_tree(
            load_balancer_id=self.load_balancer_id,
            listener_ids=[self.listener_id, new_listener_id])
        listeners = self.listeners_client.list_listeners()
        self.assertEqual(len(listeners), 2)
        self._prep_list_comparison(self.listener, listeners)
        self._prep_list_comparison(new_listener, listeners)
        self.assertIn(self.listener, listeners)
        self.assertIn(new_listener, listeners)
        self.assertNotEqual(self.listener, new_listener)

    @decorators.attr(type='smoke')
    def test_create_listener(self):
        """Test create listener"""
        create_new_listener_kwargs = self.create_listener_kwargs
        create_new_listener_kwargs['protocol_port'] = 8081
        new_listener = self._create_listener(
            **create_new_listener_kwargs)
        new_listener_id = new_listener['id']
        self.addCleanup(self._delete_listener, new_listener_id)
        self._check_status_tree(
            load_balancer_id=self.load_balancer_id,
            listener_ids=[self.listener_id, new_listener_id])
        listener = self.listeners_client.get_listener(
            new_listener_id)
        self._test_provisioning_status_if_exists(new_listener, listener)
        self.assertEqual(new_listener, listener)
        self._test_provisioning_status_if_exists(self.listener, new_listener)
        self.assertNotEqual(self.listener, new_listener)

    @decorators.attr(type='negative')
    def test_create_listener_missing_field_loadbalancer(self):
        """Test create listener with a missing required field loadbalancer"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          protocol_port=self.port,
                          protocol=self.listener_protocol)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_missing_field_protocol(self):
        """Test create listener with a missing required field protocol"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol_port=self.port)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_missing_field_protocol_port(self):
        """Test create listener with a missing required field protocol_port"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol=self.listener_protocol)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    def test_create_listener_missing_admin_state_up(self):
        """Test create listener with a missing admin_state_up field"""
        create_new_listener_kwargs = self.create_listener_kwargs
        create_new_listener_kwargs['protocol_port'] = 8081
        new_listener = self._create_listener(
            **create_new_listener_kwargs)
        new_listener_id = new_listener['id']
        self.addCleanup(self._delete_listener, new_listener_id)
        self._check_status_tree(
            load_balancer_id=self.load_balancer_id,
            listener_ids=[self.listener_id, new_listener_id])
        listener = self.listeners_client.get_listener(
            new_listener_id)
        self._test_provisioning_status_if_exists(new_listener, listener)
        self.assertEqual(new_listener, listener)
        self.assertTrue(new_listener['admin_state_up'])

    @decorators.attr(type='negative')
    def test_create_listener_invalid_load_balancer_id(self):
        """Test create listener with an invalid load_balancer_id"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id="234*",
                          protocol_port=self.port,
                          protocol=self.listener_protocol)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_invalid_protocol(self):
        """Test create listener with an invalid protocol"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol_port=self.port,
                          protocol="UDP")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_invalid_protocol_port(self):
        """Test create listener with an invalid protocol_port"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol_port="9999999",
                          protocol=self.listener_protocol)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_invalid_admin_state_up(self):
        """Test update listener with an invalid admin_state_up"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          protocol_port=self.port,
                          protocol=self.listener_protocol,
                          admin_state_up="abc123")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_listener_invalid_tenant_id(self):
        """Test create listener with an invalid tenant id"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol_port=self.port,
                          protocol=self.listener_protocol,
                          tenant_id="&^%123")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_invalid_name(self):
        """Test create listener with an invalid name"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol_port=self.port,
                          protocol=self.listener_protocol,
                          name='a' * 256)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_invalid_description(self):
        """Test create listener with an invalid description"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol_port=self.port,
                          protocol=self.listener_protocol,
                          description='a' * 256)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_invalid_connection_limit(self):
        """Test create listener with an invalid value for connection
        _limit field
        """
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol_port=self.port,
                          protocol=self.listener_protocol,
                          connection_limit="&^%123")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_empty_load_balancer_id(self):
        """Test create listener with an empty load_balancer_id"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id="",
                          protocol_port=self.port,
                          protocol=self.listener_protocol)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_empty_protocol(self):
        """Test create listener with an empty protocol"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol_port=self.port,
                          protocol="")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_empty_protocol_port(self):
        """Test create listener with an empty protocol_port"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol_port="",
                          protocol=self.listener_protocol)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_empty_admin_state_up(self):
        """Test update listener with an empty  admin_state_up"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          protocol_port=self.port,
                          protocol=self.listener_protocol,
                          admin_state_up="")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_listener_empty_tenant_id(self):
        """Test create listener with an empty tenant id"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol_port=self.port,
                          protocol=self.listener_protocol,
                          tenant_id="")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    def test_create_listener_empty_name(self):
        """Test create listener with an empty name"""
        create_new_listener_kwargs = self.create_listener_kwargs
        create_new_listener_kwargs['protocol_port'] = 8081
        create_new_listener_kwargs['name'] = ""
        new_listener = self._create_listener(
            **create_new_listener_kwargs)
        new_listener_id = new_listener['id']
        self.addCleanup(self._delete_listener, new_listener_id)
        self._check_status_tree(
            load_balancer_id=self.load_balancer_id,
            listener_ids=[self.listener_id, new_listener_id])
        listener = self.listeners_client.get_listener(
            new_listener_id)
        self._test_provisioning_status_if_exists(new_listener, listener)
        self.assertEqual(new_listener, listener)

    def test_create_listener_empty_description(self):
        """Test create listener with an empty description"""
        create_new_listener_kwargs = self.create_listener_kwargs
        create_new_listener_kwargs['protocol_port'] = 8081
        create_new_listener_kwargs['description'] = ""
        new_listener = self._create_listener(
            **create_new_listener_kwargs)
        new_listener_id = new_listener['id']
        self.addCleanup(self._delete_listener, new_listener_id)
        self._check_status_tree(
            load_balancer_id=self.load_balancer_id,
            listener_ids=[self.listener_id, new_listener_id])
        listener = self.listeners_client.get_listener(
            new_listener_id)
        self._test_provisioning_status_if_exists(new_listener, listener)
        self.assertEqual(new_listener, listener)

    @decorators.attr(type='negative')
    def test_create_listener_empty_connection_limit(self):
        """Test create listener with an empty connection
        _limit field
        """
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          loadbalancer_id=self.load_balancer_id,
                          protocol_port=self.port,
                          protocol=self.listener_protocol,
                          connection_limit="")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_create_listener_incorrect_attribute(self):
        """Test create a listener with an extra, incorrect field"""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          incorrect_attribute="incorrect_attribute",
                          **self.create_listener_kwargs)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='smoke')
    def test_update_listener(self):
        """Test update listener"""
        self._update_listener(self.listener_id,
                              name='new_name')
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])
        listener = self.listeners_client.get_listener(
            self.listener_id)
        self.assertEqual(listener.get('name'), 'new_name')

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_update_listener_invalid_tenant_id(self):
        """Test update listener with an invalid tenant id"""
        self.assertRaises(ex.BadRequest,
                          self._update_listener,
                          listener_id=self.listener_id,
                          tenant_id="&^%123")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_update_listener_invalid_admin_state_up(self):
        """Test update a listener with an invalid admin_state_up"""
        self.assertRaises(ex.BadRequest,
                          self._update_listener,
                          listener_id=self.listener_id,
                          admin_state_up="$23")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_update_listener_invalid_name(self):
        """Test update a listener with an invalid name"""
        self.assertRaises(ex.BadRequest,
                          self._update_listener,
                          listener_id=self.listener_id,
                          name='a' * 256)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_update_listener_invalid_description(self):
        """Test update a listener with an invalid description"""
        self.assertRaises(ex.BadRequest,
                          self._update_listener,
                          listener_id=self.listener_id,
                          description='a' * 256)
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_update_listener_invalid_connection_limit(self):
        """Test update a listener with an invalid connection_limit"""
        self.assertRaises(ex.BadRequest,
                          self._update_listener,
                          listener_id=self.listener_id,
                          connection_limit="$23")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_update_listener_incorrect_attribute(self):
        """Test update a listener with an extra, incorrect field"""
        self.assertRaises(ex.BadRequest,
                          self._update_listener,
                          listener_id=self.listener_id,
                          name="listener_name123",
                          description="listener_description123",
                          admin_state_up=True,
                          connection_limit=10,
                          vip_subnet_id="123321123")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    def test_update_listener_missing_name(self):
        """Test update listener with a missing name"""
        old_listener = self.listeners_client.get_listener(
            self.listener_id)
        old_name = old_listener['name']
        self._update_listener(self.listener_id,
                              description='updated')
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])
        listener = self.listeners_client.get_listener(
            self.listener_id)
        self.assertEqual(listener.get('name'), old_name)

    def test_update_listener_missing_description(self):
        """Test update listener with a missing description"""
        old_listener = self.listeners_client.get_listener(
            self.listener_id)
        old_description = old_listener['description']
        self._update_listener(self.listener_id,
                              name='updated_name')
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])
        listener = self.listeners_client.get_listener(
            self.listener_id)
        self.assertEqual(listener.get('description'), old_description)

    def test_update_listener_missing_admin_state_up(self):
        """Test update listener with a missing admin_state_up"""
        old_listener = self.listeners_client.get_listener(
            self.listener_id)
        old_admin_state_up = old_listener['admin_state_up']
        self._update_listener(self.listener_id,
                              name='updated_name')
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])
        listener = self.listeners_client.get_listener(
            self.listener_id)
        self.assertEqual(listener.get('admin_state_up'), old_admin_state_up)

    def test_update_listener_missing_connection_limit(self):
        """Test update listener with a missing connection_limit"""
        old_listener = self.listeners_client.get_listener(
            self.listener_id)
        old_connection_limit = old_listener['connection_limit']
        self._update_listener(self.listener_id,
                              name='updated_name')
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])
        listener = self.listeners_client.get_listener(
            self.listener_id)
        self.assertEqual(listener.get('connection_limit'),
                         old_connection_limit)

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_update_listener_empty_tenant_id(self):
        """Test update listener with an empty tenant id"""
        self.assertRaises(ex.BadRequest,
                          self._update_listener,
                          listener_id=self.listener_id,
                          tenant_id="")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='negative')
    def test_update_listener_empty_admin_state_up(self):
        """Test update a listener with an empty admin_state_up"""
        self.assertRaises(ex.BadRequest,
                          self._update_listener,
                          listener_id=self.listener_id,
                          admin_state_up="")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    def test_update_listener_empty_name(self):
        """Test update a listener with an empty name"""
        self._update_listener(self.listener_id,
                              name="")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])
        listener = self.listeners_client.get_listener(
            self.listener_id)
        self.assertEqual(listener.get('name'), "")

    def test_update_listener_empty_description(self):
        """Test update a listener with an empty description"""
        self._update_listener(self.listener_id,
                              description="")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])
        listener = self.listeners_client.get_listener(
            self.listener_id)
        self.assertEqual(listener.get('description'), "")

    @decorators.attr(type='negative')
    def test_update_listener_empty_connection_limit(self):
        """Test update a listener with an empty connection_limit"""
        self.assertRaises(ex.BadRequest,
                          self._update_listener,
                          listener_id=self.listener_id,
                          connection_limit="")
        self._check_status_tree(load_balancer_id=self.load_balancer_id,
                                listener_ids=[self.listener_id])

    @decorators.attr(type='smoke')
    def test_delete_listener(self):
        """Test delete listener"""
        create_new_listener_kwargs = self.create_listener_kwargs
        create_new_listener_kwargs['protocol_port'] = 8083
        new_listener = self._create_listener(**create_new_listener_kwargs)
        new_listener_id = new_listener['id']
        self._check_status_tree(
            load_balancer_id=self.load_balancer_id,
            listener_ids=[self.listener_id, new_listener_id])
        listener = self.listeners_client.get_listener(
            new_listener_id)
        self._test_provisioning_status_if_exists(new_listener, listener)
        self.assertEqual(new_listener, listener)
        self._test_provisioning_status_if_exists(self.listener, new_listener)
        self.assertNotEqual(self.listener, new_listener)
        self._delete_listener(new_listener_id)
        self.assertRaises(ex.NotFound,
                          self.listeners_client.get_listener,
                          new_listener_id)



# Copyright 2017, Rackspace US, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron_lib import constants
from neutron_lib import context
import webob.exc

from neutron_lbaas._i18n import _
from neutron_lbaas.db.loadbalancer import models
from neutron_lbaas.extensions import healthmonitor_max_retries_down
from neutron_lbaas.extensions import l7
from neutron_lbaas.extensions import lb_graph
from neutron_lbaas.extensions import lb_network_vip
from neutron_lbaas.extensions import loadbalancerv2
from neutron_lbaas.extensions import sharedpools
from neutron_lbaas.services.loadbalancer import constants as lb_const


class ExtendedPluginAwareExtensionManager(object):
    def __init__(self, extension_aliases):
        self.extension_aliases = extension_aliases

    def get_resources(self):
        extensions_list = []
        if 'shared_pools' in self.extension_aliases:
            extensions_list.append(sharedpools)
        if 'l7' in self.extension_aliases:
            extensions_list.append(l7)
        if 'lb-graph' in self.extension_aliases:
            extensions_list.append(lb_graph)
        if 'lb_network_vip' in self.extension_aliases:
            extensions_list.append(lb_network_vip)
        if 'hm_max_retries_down' in self.extension_aliases:
            extensions_list.append(healthmonitor_max_retries_down)
        for extension in extensions_list:
            if 'RESOURCE_ATTRIBUTE_MAP' in extension.__dict__:
                loadbalancerv2.RESOURCE_ATTRIBUTE_MAP.update(
                    extension.RESOURCE_ATTRIBUTE_MAP)
            if 'SUB_RESOURCE_ATTRIBUTE_MAP' in extension.__dict__:
                loadbalancerv2.SUB_RESOURCE_ATTRIBUTE_MAP.update(
                    extension.SUB_RESOURCE_ATTRIBUTE_MAP)
            if 'EXTENDED_ATTRIBUTES_2_0' in extension.__dict__:
                for key in loadbalancerv2.RESOURCE_ATTRIBUTE_MAP.keys():
                    loadbalancerv2.RESOURCE_ATTRIBUTE_MAP[key].update(
                        extension.EXTENDED_ATTRIBUTES_2_0.get(key, {}))
        return loadbalancerv2.Loadbalancerv2.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class LbaasTestMixin(object):
    resource_keys = list(loadbalancerv2.RESOURCE_ATTRIBUTE_MAP.keys())
    resource_keys.extend(l7.RESOURCE_ATTRIBUTE_MAP.keys())
    resource_keys.extend(lb_graph.RESOURCE_ATTRIBUTE_MAP.keys())
    resource_keys.extend(lb_network_vip.EXTENDED_ATTRIBUTES_2_0.keys())
    resource_keys.extend(healthmonitor_max_retries_down.
                         EXTENDED_ATTRIBUTES_2_0.keys())
    resource_prefix_map = dict(
        (k, loadbalancerv2.LOADBALANCERV2_PREFIX)
        for k in resource_keys)

    def _get_loadbalancer_optional_args(self):
        return ('description', 'vip_address', 'admin_state_up', 'name',
                'listeners', 'vip_network_id', 'vip_subnet_id')

    def _create_loadbalancer(self, fmt, subnet_id,
                             expected_res_status=None, **kwargs):
        data = {'loadbalancer': {'vip_subnet_id': subnet_id,
                                 'tenant_id': self._tenant_id}}
        args = self._get_loadbalancer_optional_args()
        for arg in args:
            if arg in kwargs:
                if kwargs[arg] is not None:
                    data['loadbalancer'][arg] = kwargs[arg]
                else:
                    data['loadbalancer'].pop(arg, None)

        lb_req = self.new_create_request('loadbalancers', data, fmt)
        lb_res = lb_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, lb_res.status_int)

        return lb_res

    def _create_graph(self, fmt, subnet_id, expected_res_status=None,
                      **kwargs):
        data = {'vip_subnet_id': subnet_id, 'tenant_id': self._tenant_id}
        args = self._get_loadbalancer_optional_args()
        for arg in args:
            if arg in kwargs and kwargs[arg] is not None:
                data[arg] = kwargs[arg]

        data = {'graph': {'loadbalancer': data, 'tenant_id': self._tenant_id}}
        lb_req = self.new_create_request('graphs', data, fmt)
        lb_res = lb_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, lb_res.status_int)

        return lb_res

    def _get_listener_optional_args(self):
        return ('name', 'description', 'default_pool_id', 'loadbalancer_id',
                'connection_limit', 'admin_state_up',
                'default_tls_container_ref', 'sni_container_refs')

    def _create_listener(self, fmt, protocol, protocol_port,
                         loadbalancer_id=None, default_pool_id=None,
                         expected_res_status=None, **kwargs):
        data = {'listener': {'protocol': protocol,
                             'protocol_port': protocol_port,
                             'tenant_id': self._tenant_id}}
        if loadbalancer_id:
            data['listener']['loadbalancer_id'] = loadbalancer_id
        if default_pool_id:
            data['listener']['default_pool_id'] = default_pool_id

        args = self._get_listener_optional_args()
        for arg in args:
            if arg in kwargs and kwargs[arg] is not None:
                data['listener'][arg] = kwargs[arg]

        listener_req = self.new_create_request('listeners', data, fmt)
        listener_res = listener_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, listener_res.status_int)

        return listener_res

    def _get_pool_optional_args(self):
        return 'name', 'description', 'admin_state_up', 'session_persistence'

    def _create_pool(self, fmt, protocol, lb_algorithm, listener_id=None,
                     loadbalancer_id=None, expected_res_status=None, **kwargs):
        data = {'pool': {'protocol': protocol,
                         'lb_algorithm': lb_algorithm,
                         'tenant_id': self._tenant_id}}
        if listener_id:
            data['pool']['listener_id'] = listener_id
        if loadbalancer_id:
            data['pool']['loadbalancer_id'] = loadbalancer_id

        args = self._get_pool_optional_args()
        for arg in args:
            if arg in kwargs and kwargs[arg] is not None:
                data['pool'][arg] = kwargs[arg]

        pool_req = self.new_create_request('pools', data, fmt)
        pool_res = pool_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, pool_res.status_int)

        return pool_res

    def _get_member_optional_args(self):
        return 'weight', 'admin_state_up', 'name'

    def _create_member(self, fmt, pool_id, address, protocol_port, subnet_id,
                       expected_res_status=None, **kwargs):
        data = {'member': {'address': address,
                           'protocol_port': protocol_port,
                           'subnet_id': subnet_id,
                           'tenant_id': self._tenant_id}}

        args = self._get_member_optional_args()
        for arg in args:
            if arg in kwargs and kwargs[arg] is not None:
                data['member'][arg] = kwargs[arg]
        member_req = self.new_create_request('pools',
                                             data,
                                             fmt=fmt,
                                             id=pool_id,
                                             subresource='members')
        member_res = member_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, member_res.status_int)

        return member_res

    def _get_healthmonitor_optional_args(self):
        return ('weight', 'admin_state_up', 'expected_codes', 'url_path',
                'http_method', 'name', 'max_retries_down')

    def _create_healthmonitor(self, fmt, pool_id, type, delay, timeout,
                              max_retries, expected_res_status=None, **kwargs):
        data = {'healthmonitor': {'type': type,
                                  'delay': delay,
                                  'timeout': timeout,
                                  'max_retries': max_retries,
                                  'pool_id': pool_id,
                                  'tenant_id': self._tenant_id}}

        args = self._get_healthmonitor_optional_args()
        for arg in args:
            if arg in kwargs and kwargs[arg] is not None:
                data['healthmonitor'][arg] = kwargs[arg]

        hm_req = self.new_create_request('healthmonitors', data, fmt=fmt)
        hm_res = hm_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, hm_res.status_int)

        return hm_res

    def _add_optional_args(self, optional_args, data, **kwargs):
        for arg in optional_args:
            if arg in kwargs and kwargs[arg] is not None:
                data[arg] = kwargs[arg]

    def _get_l7policy_optional_args(self):
        return ('name', 'description', 'redirect_pool_id',
                'redirect_url', 'admin_state_up', 'position')

    def _create_l7policy(self, fmt, listener_id, action,
                         expected_res_status=None, **kwargs):
        data = {'l7policy': {'listener_id': listener_id,
                             'action': action,
                             'tenant_id': self._tenant_id}}

        optional_args = self._get_l7policy_optional_args()
        self._add_optional_args(optional_args, data['l7policy'], **kwargs)

        l7policy_req = self.new_create_request('l7policies', data, fmt)
        l7policy_res = l7policy_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(l7policy_res.status_int, expected_res_status)

        return l7policy_res

    def _get_l7rule_optional_args(self):
        return ('invert', 'key', 'admin_state_up')

    def _create_l7policy_rule(self, fmt, l7policy_id, type, compare_type,
                              value, expected_res_status=None, **kwargs):
        data = {'rule': {'type': type,
                         'compare_type': compare_type,
                         'value': value,
                         'tenant_id': self._tenant_id}}

        optional_args = self._get_l7rule_optional_args()
        self._add_optional_args(optional_args, data['rule'], **kwargs)

        rule_req = self.new_create_request('l7policies', data, fmt,
                                           id=l7policy_id,
                                           subresource='rules')
        rule_res = rule_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(rule_res.status_int, expected_res_status)

        return rule_res

    @contextlib.contextmanager
    def loadbalancer(self, fmt=None, subnet=None, no_delete=False, **kwargs):
        if not fmt:
            fmt = self.fmt

        with test_db_base_plugin_v2.optional_ctx(
                subnet, self.subnet) as tmp_subnet:

            res = self._create_loadbalancer(fmt,
                                            tmp_subnet['subnet']['id'],
                                            **kwargs)
            if res.status_int >= webob.exc.HTTPClientError.code:
                exc = webob.exc.HTTPClientError(
                    explanation=_("Unexpected error code: %s") %
                    res.status_int)
                exc.code = res.status_int
                exc.status_code = res.status_int
                raise exc
            lb = self.deserialize(fmt or self.fmt, res)
            yield lb
            if not no_delete:
                self._delete('loadbalancers', lb['loadbalancer']['id'])

    @contextlib.contextmanager
    def graph(self, fmt=None, subnet=None, no_delete=False, **kwargs):
        if not fmt:
            fmt = self.fmt

        with test_db_base_plugin_v2.optional_ctx(
                subnet, self.subnet) as tmp_subnet:

            res = self._create_graph(fmt, tmp_subnet['subnet']['id'],
                                     **kwargs)
            if res.status_int >= webob.exc.HTTPClientError.code:
                exc = webob.exc.HTTPClientError(
                    explanation=_("Unexpected error code: %s") %
                    res.status_int
                )
                exc.code = res.status_int
                exc.status_code = res.status_int
                raise exc
            graph = self.deserialize(fmt or self.fmt, res)
            yield graph
            if not no_delete:
                # delete loadbalancer children if this was a loadbalancer
                # graph create call
                lb = graph['graph']['loadbalancer']
                for listener in lb.get('listeners', []):
                    pool = listener.get('default_pool')
                    if pool:
                        hm = pool.get('healthmonitor')
                        if hm:
                            self._delete('healthmonitors', hm['id'])
                        members = pool.get('members', [])
                        for member in members:
                            self._delete('pools', pool['id'],
                                         subresource='members',
                                         sub_id=member['id'])
                        self._delete('pools', pool['id'])
                    policies = listener.get('l7policies', [])
                    for policy in policies:
                        r_pool = policy.get('redirect_pool')
                        if r_pool:
                            r_hm = r_pool.get('healthmonitor')
                            if r_hm:
                                self._delete('healthmonitors', r_hm['id'])
                            r_members = r_pool.get('members', [])
                            for r_member in r_members:
                                self._delete('pools', r_pool['id'],
                                             subresource='members',
                                             sub_id=r_member['id'])
                            self._delete('pools', r_pool['id'])
                        self._delete('l7policies', policy['id'])
                    self._delete('listeners', listener['id'])
                self._delete('loadbalancers', lb['id'])

    @contextlib.contextmanager
    def listener(self, fmt=None, protocol='HTTP', loadbalancer_id=None,
                 protocol_port=80, default_pool_id=None, no_delete=False,
                 **kwargs):
        if not fmt:
            fmt = self.fmt

        if loadbalancer_id and default_pool_id:
            res = self._create_listener(fmt, protocol, protocol_port,
                                        loadbalancer_id=loadbalancer_id,
                                        default_pool_id=default_pool_id,
                                        **kwargs)
        elif loadbalancer_id:
            res = self._create_listener(fmt, protocol, protocol_port,
                                        loadbalancer_id=loadbalancer_id,
                                        **kwargs)
        else:
            res = self._create_listener(fmt, protocol, protocol_port,
                                        default_pool_id=default_pool_id,
                                        **kwargs)
        if res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(
                explanation=_("Unexpected error code: %s") % res.status_int
            )

        listener = self.deserialize(fmt or self.fmt, res)
        yield listener
        if not no_delete:
            self._delete('listeners', listener['listener']['id'])

    @contextlib.contextmanager
    def pool(self, fmt=None, protocol='HTTP', lb_algorithm='ROUND_ROBIN',
             no_delete=False, listener_id=None,
             loadbalancer_id=None, **kwargs):
        if not fmt:
            fmt = self.fmt

        if listener_id and loadbalancer_id:
            res = self._create_pool(fmt,
                                    protocol=protocol,
                                    lb_algorithm=lb_algorithm,
                                    listener_id=listener_id,
                                    loadbalancer_id=loadbalancer_id,
                                    **kwargs)
        elif listener_id:
            res = self._create_pool(fmt,
                                    protocol=protocol,
                                    lb_algorithm=lb_algorithm,
                                    listener_id=listener_id,
                                    **kwargs)
        else:
            res = self._create_pool(fmt,
                                    protocol=protocol,
                                    lb_algorithm=lb_algorithm,
                                    loadbalancer_id=loadbalancer_id,
                                    **kwargs)
        if res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(
                explanation=_("Unexpected error code: %s") % res.status_int
            )

        pool = self.deserialize(fmt or self.fmt, res)
        yield pool
        if not no_delete:
            self._delete('pools', pool['pool']['id'])

    @contextlib.contextmanager
    def member(self, fmt=None, pool_id='pool1id', address='127.0.0.1',
               protocol_port=80, subnet=None, no_delete=False,
               **kwargs):
        if not fmt:
            fmt = self.fmt
        subnet = subnet or self.test_subnet
        with test_db_base_plugin_v2.optional_ctx(
                subnet, self.subnet) as tmp_subnet:

            res = self._create_member(fmt,
                                      pool_id=pool_id,
                                      address=address,
                                      protocol_port=protocol_port,
                                      subnet_id=tmp_subnet['subnet']['id'],
                                      **kwargs)
            if res.status_int >= webob.exc.HTTPClientError.code:
                raise webob.exc.HTTPClientError(
                    explanation=_("Unexpected error code: %s") % res.status_int
                )

            member = self.deserialize(fmt or self.fmt, res)
        yield member
        if not no_delete:
            self._delete('pools', id=pool_id, subresource='members',
                         sub_id=member['member']['id'])

    @contextlib.contextmanager
    def healthmonitor(self, fmt=None, pool_id='pool1id', type='TCP', delay=1,
                      timeout=1, max_retries=2, no_delete=False, **kwargs):
        if not fmt:
            fmt = self.fmt

        res = self._create_healthmonitor(fmt,
                                         pool_id=pool_id,
                                         type=type,
                                         delay=delay,
                                         timeout=timeout,
                                         max_retries=max_retries,
                                         **kwargs)
        if res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(
                explanation=_("Unexpected error code: %s") % res.status_int
            )

        healthmonitor = self.deserialize(fmt or self.fmt, res)
        yield healthmonitor
        if not no_delete:
            del_req = self.new_delete_request(
                'healthmonitors', fmt=fmt,
                id=healthmonitor['healthmonitor']['id'])
            del_res = del_req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, del_res.status_int)

    @contextlib.contextmanager
    def l7policy(self, listener_id, fmt=None,
                 action=lb_const.L7_POLICY_ACTION_REJECT,
                 no_delete=False, **kwargs):
        if not fmt:
            fmt = self.fmt

        res = self._create_l7policy(fmt,
                                    listener_id=listener_id,
                                    action=action,
                                    **kwargs)
        if res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(
                explanation=_("Unexpected error code: %s") % res.status_int
            )

        l7policy = self.deserialize(fmt or self.fmt, res)
        yield l7policy
        if not no_delete:
            self.plugin.db.update_status(context.get_admin_context(),
                                         models.L7Policy,
                                         l7policy['l7policy']['id'],
                                         constants.ACTIVE)
            del_req = self.new_delete_request(
                'l7policies',
                fmt=fmt,
                id=l7policy['l7policy']['id'])
            del_res = del_req.get_response(self.ext_api)
            self.assertEqual(del_res.status_int,
                             webob.exc.HTTPNoContent.code)

    @contextlib.contextmanager
    def l7policy_rule(self, l7policy_id, fmt=None, value='value1',
                      type=lb_const.L7_RULE_TYPE_HOST_NAME,
                      compare_type=lb_const.L7_RULE_COMPARE_TYPE_EQUAL_TO,
                      no_delete=False, **kwargs):
        if not fmt:
            fmt = self.fmt
        res = self._create_l7policy_rule(fmt,
                                         l7policy_id=l7policy_id,
                                         type=type,
                                         compare_type=compare_type,
                                         value=value,
                                         **kwargs)
        if res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(
                explanation=_("Unexpected error code: %s") % res.status_int
            )

        rule = self.deserialize(fmt or self.fmt, res)
        yield rule
        if not no_delete:
            self.plugin.db.update_status(context.get_admin_context(),
                                         models.L7Rule,
                                         rule['rule']['id'],
                                         constants.ACTIVE)
            del_req = self.new_delete_request(
                'l7policies',
                fmt=fmt,
                id=l7policy_id,
                subresource='rules',
                sub_id=rule['rule']['id'])
            del_res = del_req.get_response(self.ext_api)
            self.assertEqual(del_res.status_int,
                             webob.exc.HTTPNoContent.code)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import jinja2
from neutron_lib.utils import file as file_utils
from oslo_config import cfg
import six

from neutron_lib import constants as nl_constants

from neutron_lbaas._i18n import _
from neutron_lbaas.common import cert_manager
from neutron_lbaas.common.tls_utils import cert_parser
from neutron_lbaas.services.loadbalancer import constants
from neutron_lbaas.services.loadbalancer import data_models

CERT_MANAGER_PLUGIN = cert_manager.get_backend()

PROTOCOL_MAP = {
    constants.PROTOCOL_TCP: 'tcp',
    constants.PROTOCOL_HTTP: 'http',
    constants.PROTOCOL_HTTPS: 'tcp',
    constants.PROTOCOL_TERMINATED_HTTPS: 'http'
}

BALANCE_MAP = {
    constants.LB_METHOD_ROUND_ROBIN: 'roundrobin',
    constants.LB_METHOD_LEAST_CONNECTIONS: 'leastconn',
    constants.LB_METHOD_SOURCE_IP: 'source'
}

STATS_MAP = {
    constants.STATS_ACTIVE_CONNECTIONS: 'scur',
    constants.STATS_MAX_CONNECTIONS: 'smax',
    constants.STATS_CURRENT_SESSIONS: 'scur',
    constants.STATS_MAX_SESSIONS: 'smax',
    constants.STATS_TOTAL_CONNECTIONS: 'stot',
    constants.STATS_TOTAL_SESSIONS: 'stot',
    constants.STATS_IN_BYTES: 'bin',
    constants.STATS_OUT_BYTES: 'bout',
    constants.STATS_CONNECTION_ERRORS: 'econ',
    constants.STATS_RESPONSE_ERRORS: 'eresp'
}

MEMBER_STATUSES = nl_constants.ACTIVE_PENDING_STATUSES + (
    nl_constants.INACTIVE,)

TEMPLATES_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'templates/'))
JINJA_ENV = None

jinja_opts = [
    cfg.StrOpt(
        'jinja_config_template',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        default=os.path.join(
            TEMPLATES_DIR,
            'haproxy.loadbalancer.j2'),
        help=_('Jinja template file for haproxy configuration'))
]

cfg.CONF.register_opts(jinja_opts, 'haproxy')


def save_config(conf_path, loadbalancer, socket_path, user_group,
                haproxy_base_dir):
    """Convert a logical configuration to the HAProxy version.

    :param conf_path: location of Haproxy configuration
    :param loadbalancer: the load balancer object
    :param socket_path: location of haproxy socket data
    :param user_group: user group
    :param haproxy_base_dir: location of the instances state data
    """
    config_str = render_loadbalancer_obj(loadbalancer,
                                         user_group,
                                         socket_path,
                                         haproxy_base_dir)
    file_utils.replace_file(conf_path, config_str)


def _get_template():
    """Retrieve Jinja template

    :returns: Jinja template
    """
    global JINJA_ENV
    if not JINJA_ENV:
        template_loader = jinja2.FileSystemLoader(
            searchpath=os.path.dirname(cfg.CONF.haproxy.jinja_config_template))
        JINJA_ENV = jinja2.Environment(
            loader=template_loader, trim_blocks=True, lstrip_blocks=True)
    return JINJA_ENV.get_template(os.path.basename(
        cfg.CONF.haproxy.jinja_config_template))


def _store_listener_crt(haproxy_base_dir, listener, cert):
    """Store TLS certificate

    :param haproxy_base_dir: location of the instances state data
    :param listener: the listener object
    :param cert: the TLS certificate
    :returns: location of the stored certificate
    """
    cert_path = _retrieve_crt_path(haproxy_base_dir, listener,
                                   cert.primary_cn)
    # build a string that represents the pem file to be saved
    pem = _build_pem(cert)
    file_utils.replace_file(cert_path, pem)
    return cert_path


def _retrieve_crt_path(haproxy_base_dir, listener, primary_cn):
    """Retrieve TLS certificate location

    :param haproxy_base_dir: location of the instances state data
    :param listener: the listener object
    :param primary_cn: primary_cn used for identifying TLS certificate
    :returns: TLS certificate location
    """
    confs_dir = os.path.abspath(os.path.normpath(haproxy_base_dir))
    confs_path = os.path.join(confs_dir, listener.id)
    if haproxy_base_dir and listener.id:
        if not os.path.isdir(confs_path):
            os.makedirs(confs_path, 0o755)
        return os.path.join(
            confs_path, '{0}.pem'.format(primary_cn))


def _process_tls_certificates(listener):
    """Processes TLS data from the listener.

    Converts and uploads PEM data to the Amphora API

    :param listener: the listener object
    :returns: TLS_CERT and SNI_CERTS
    """
    cert_mgr = CERT_MANAGER_PLUGIN.CertManager()

    tls_cert = None
    sni_certs = []
    # Retrieve, map and store default TLS certificate
    if listener.default_tls_container_id:
        tls_cert = _map_cert_tls_container(
            cert_mgr.get_cert(
                project_id=listener.tenant_id,
                cert_ref=listener.default_tls_container_id,
                resource_ref=cert_mgr.get_service_url(
                    listener.loadbalancer_id),
                check_only=True
            )
        )
    if listener.sni_containers:
        # Retrieve, map and store SNI certificates
        for sni_cont in listener.sni_containers:
            cert_container = _map_cert_tls_container(
                cert_mgr.get_cert(
                    project_id=listener.tenant_id,
                    cert_ref=sni_cont.tls_container_id,
                    resource_ref=cert_mgr.get_service_url(
                        listener.loadbalancer_id),
                    check_only=True
                )
            )
            sni_certs.append(cert_container)

    return {'tls_cert': tls_cert, 'sni_certs': sni_certs}


def _get_primary_cn(tls_cert):
    """Retrieve primary cn for TLS certificate

    :param tls_cert: the TLS certificate
    :returns: primary cn of the TLS certificate
    """
    return cert_parser.get_host_names(tls_cert)['cn']


def _map_cert_tls_container(cert):
    """Map cert data to TLS data model

    :param cert: TLS certificate
    :returns: mapped TLSContainer object
    """
    certificate = cert.get_certificate()
    pkey = cert_parser.dump_private_key(cert.get_private_key(),
                                        cert.get_private_key_passphrase())
    return data_models.TLSContainer(
        primary_cn=_get_primary_cn(certificate),
        private_key=pkey,
        certificate=certificate,
        intermediates=cert.get_intermediates())


def _build_pem(tls_cert):
    """Generate PEM encoded TLS certificate data

    :param tls_cert: TLS certificate
    :returns: PEm encoded certificate data
    """
    pem = ()
    if tls_cert.intermediates:
        for c in tls_cert.intermediates:
            pem = pem + (c,)
    if tls_cert.certificate:
        pem = pem + (tls_cert.certificate,)
    if tls_cert.private_key:
        pem = pem + (tls_cert.private_key,)
    return "\n".join(pem)


def render_loadbalancer_obj(loadbalancer, user_group, socket_path,
                            haproxy_base_dir):
    """Renders load balancer object

    :param loadbalancer: the load balancer object
    :param user_group: the user group
    :param socket_path: location of the instances socket data
    :param haproxy_base_dir:  location of the instances state data
    :returns: rendered load balancer configuration
    """
    loadbalancer = _transform_loadbalancer(loadbalancer, haproxy_base_dir)
    return _get_template().render({'loadbalancer': loadbalancer,
                                   'user_group': user_group,
                                   'stats_sock': socket_path},
                                  constants=constants)


def _transform_loadbalancer(loadbalancer, haproxy_base_dir):
    """Transforms load balancer object

    :param loadbalancer: the load balancer object
    :param haproxy_base_dir: location of the instances state data
    :returns: dictionary of transformed load balancer values
    """
    listeners = [_transform_listener(x, haproxy_base_dir)
        for x in loadbalancer.listeners if x.admin_state_up]
    pools = [_transform_pool(x) for x in loadbalancer.pools]
    connection_limit = _compute_global_connection_limit(listeners)
    return {
        'id': loadbalancer.id,
        'vip_address': loadbalancer.vip_address,
        'connection_limit': connection_limit,
        'listeners': listeners,
        'pools': pools
    }


def _compute_global_connection_limit(listeners):
    # NOTE(dlundquist): HAProxy has a global default connection limit
    # of 2000, so we will include 2000 connections for each listener
    # without a connection limit specified. This way we provide the
    # same behavior as a default haproxy configuration without
    # connection limit specified in the case of a single load balancer.
    return sum([x.get('connection_limit', 2000) for x in listeners])


def _transform_listener(listener, haproxy_base_dir):
    """Transforms listener object

    :param listener: the listener object
    :param haproxy_base_dir: location of the instances state data
    :returns: dictionary of transformed listener values
    """
    data_dir = os.path.join(haproxy_base_dir, listener.id)
    ret_value = {
        'id': listener.id,
        'protocol_port': listener.protocol_port,
        'protocol_mode': PROTOCOL_MAP[listener.protocol],
        'protocol': listener.protocol
    }
    if listener.connection_limit and listener.connection_limit > -1:
        ret_value['connection_limit'] = listener.connection_limit
    if listener.default_pool:
        ret_value['default_pool'] = _transform_pool(listener.default_pool)

    # Process and store certificates
    certs = _process_tls_certificates(listener)
    if listener.default_tls_container_id:
        ret_value['default_tls_path'] = _store_listener_crt(
            haproxy_base_dir, listener, certs['tls_cert'])
    if listener.sni_containers:
        for c in certs['sni_certs']:
            _store_listener_crt(haproxy_base_dir, listener, c)
        ret_value['crt_dir'] = data_dir
    return ret_value


def _transform_pool(pool):
    """Transforms pool object

    :param pool: the pool object
    :returns: dictionary of transformed pool values
    """
    ret_value = {
        'id': pool.id,
        'protocol': PROTOCOL_MAP[pool.protocol],
        'lb_algorithm': BALANCE_MAP.get(pool.lb_algorithm, 'roundrobin'),
        'members': [],
        'health_monitor': '',
        'session_persistence': '',
        'admin_state_up': pool.admin_state_up,
        'provisioning_status': pool.provisioning_status
    }
    members = [_transform_member(x)
               for x in pool.members if _include_member(x)]
    ret_value['members'] = members
    if pool.healthmonitor and pool.healthmonitor.admin_state_up:
        ret_value['health_monitor'] = _transform_health_monitor(
            pool.healthmonitor)
    if pool.session_persistence:
        ret_value['session_persistence'] = _transform_session_persistence(
            pool.session_persistence)
    return ret_value


def _transform_session_persistence(persistence):
    """Transforms session persistence object

    :param persistence: the session persistence object
    :returns: dictionary of transformed session persistence values
    """
    return {
        'type': persistence.type,
        'cookie_name': persistence.cookie_name
    }


def _transform_member(member):
    """Transforms member object

    :param member: the member object
    :returns: dictionary of transformed member values
    """
    return {
        'id': member.id,
        'address': member.address,
        'protocol_port': member.protocol_port,
        'weight': member.weight,
        'admin_state_up': member.admin_state_up,
        'subnet_id': member.subnet_id,
        'provisioning_status': member.provisioning_status
    }


def _transform_health_monitor(monitor):
    """Transforms health monitor object

    :param monitor: the health monitor object
    :returns: dictionary of transformed health monitor values
    """
    return {
        'id': monitor.id,
        'type': monitor.type,
        'delay': monitor.delay,
        'timeout': monitor.timeout,
        'max_retries': monitor.max_retries,
        'http_method': monitor.http_method,
        'url_path': monitor.url_path,
        'expected_codes': '|'.join(
            _expand_expected_codes(monitor.expected_codes)),
        'admin_state_up': monitor.admin_state_up,
    }


def _include_member(member):
    """Helper for verifying member statues

    :param member: the member object
    :returns: boolean of status check
    """
    return (member.provisioning_status in
            MEMBER_STATUSES and member.admin_state_up)


def _expand_expected_codes(codes):
    """Expand the expected code string in set of codes

    :param codes: string of status codes
    :returns: list of status codes
    """

    retval = set()
    for code in codes.replace(',', ' ').split(' '):
        code = code.strip()

        if not code:
            continue
        elif '-' in code:
            low, hi = code.split('-')[:2]
            retval.update(
                str(i) for i in six.moves.range(int(low), int(hi) + 1))
        else:
            retval.add(code)
    return retval



# Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from barbicanclient import client as barbican_client
import mock

import neutron_lbaas.common.cert_manager.barbican_cert_manager as bbq_common
import neutron_lbaas.tests.base as base


class TestBarbicanCert(base.BaseTestCase):
    def setUp(self):
        # Certificate data
        self.certificate = "My Certificate"
        self.intermediates = "My Intermediates"
        self.private_key = "My Private Key"
        self.private_key_passphrase = "My Private Key Passphrase"

        self.certificate_secret = barbican_client.secrets.Secret(
            api=mock.MagicMock(),
            payload=self.certificate
        )
        self.intermediates_secret = barbican_client.secrets.Secret(
            api=mock.MagicMock(),
            payload=self.intermediates
        )
        self.private_key_secret = barbican_client.secrets.Secret(
            api=mock.MagicMock(),
            payload=self.private_key
        )
        self.private_key_passphrase_secret = barbican_client.secrets.Secret(
            api=mock.MagicMock(),
            payload=self.private_key_passphrase
        )

        super(TestBarbicanCert, self).setUp()

    def test_barbican_cert(self):
        container = barbican_client.containers.CertificateContainer(
            api=mock.MagicMock(),
            certificate=self.certificate_secret,
            intermediates=self.intermediates_secret,
            private_key=self.private_key_secret,
            private_key_passphrase=self.private_key_passphrase_secret
        )
        # Create a cert
        cert = bbq_common.Cert(
            cert_container=container
        )

        # Validate the cert functions
        self.assertEqual(self.certificate, cert.get_certificate())
        self.assertEqual(self.intermediates, cert.get_intermediates())
        self.assertEqual(self.private_key, cert.get_private_key())
        self.assertEqual(self.private_key_passphrase,
                         cert.get_private_key_passphrase())



# Copyright 2013 New Dream Network, LLC (DreamHost)
# Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from neutron_lbaas.agent import agent_api as api
from neutron_lbaas.tests import base


class TestApiCache(base.BaseTestCase):
    def setUp(self):
        super(TestApiCache, self).setUp()

        self.api = api.LbaasAgentApi('topic', mock.sentinel.context, 'host')

    def test_init(self):
        self.assertEqual('host', self.api.host)
        self.assertEqual(mock.sentinel.context, self.api.context)

    def _test_method(self, method, **kwargs):
        add_host = ('get_ready_devices', 'plug_vip_port', 'unplug_vip_port')
        expected_kwargs = copy.copy(kwargs)
        if method in add_host:
            expected_kwargs['host'] = self.api.host

        with mock.patch.object(self.api.client, 'call') as rpc_mock, \
                mock.patch.object(self.api.client, 'prepare') as prepare_mock:
            prepare_mock.return_value = self.api.client
            rpc_mock.return_value = 'foo'
            rv = getattr(self.api, method)(**kwargs)

        self.assertEqual('foo', rv)

        prepare_args = {}
        prepare_mock.assert_called_once_with(**prepare_args)

        rpc_mock.assert_called_once_with(mock.sentinel.context, method,
                                         **expected_kwargs)

    def test_get_ready_devices(self):
        self._test_method('get_ready_devices')

    def test_get_loadbalancer(self):
        self._test_method('get_loadbalancer',
                          loadbalancer_id='loadbalancer_id')

    def test_loadbalancer_destroyed(self):
        self._test_method('loadbalancer_destroyed',
                          loadbalancer_id='loadbalancer_id')

    def test_loadbalancer_deployed(self):
        self._test_method('loadbalancer_deployed',
                          loadbalancer_id='loadbalancer_id')

    def test_update_status(self):
        self._test_method('update_status', obj_type='type', obj_id='id',
                          provisioning_status='p_status',
                          operating_status='o_status')

    def test_plug_vip_port(self):
        self._test_method('plug_vip_port', port_id='port_id')

    def test_unplug_vip_port(self):
        self._test_method('unplug_vip_port', port_id='port_id')

    def test_update_loadbalancer_stats(self):
        self._test_method('update_loadbalancer_stats', loadbalancer_id='id',
                          stats='stats')



# Copyright 2015 Rackspace US, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from stevedore import driver

CONF = cfg.CONF

CERT_MANAGER_DEFAULT = 'barbican'

cert_manager_opts = [
    cfg.StrOpt('cert_manager_type',
               default=CERT_MANAGER_DEFAULT,
               deprecated_for_removal=True,
               deprecated_since='Queens',
               deprecated_reason='The neutron-lbaas project is now '
                                 'deprecated. See: https://wiki.openstack.org/'
                                 'wiki/Neutron/LBaaS/Deprecation',
               help='Certificate Manager plugin. '
                    'Defaults to {0}.'.format(CERT_MANAGER_DEFAULT)),
    cfg.StrOpt('barbican_auth',
               default='barbican_acl_auth',
               deprecated_for_removal=True,
               deprecated_since='Queens',
               deprecated_reason='The neutron-lbaas project is now '
                                 'deprecated. See: https://wiki.openstack.org/'
                                 'wiki/Neutron/LBaaS/Deprecation',
               help='Name of the Barbican authentication method to use')
]

CONF.register_opts(cert_manager_opts, group='certificates')

_CERT_MANAGER_PLUGIN = None


def get_backend():
    global _CERT_MANAGER_PLUGIN
    if not _CERT_MANAGER_PLUGIN:
        _CERT_MANAGER_PLUGIN = driver.DriverManager(
            "neutron_lbaas.cert_manager.backend",
            cfg.CONF.certificates.cert_manager_type).driver
    return _CERT_MANAGER_PLUGIN



# Copyright 2014 A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from functools import wraps

from neutron_lib import context as ncontext
from oslo_utils import excutils

from neutron_lbaas.common import exceptions
from neutron_lbaas.db.loadbalancer import models
from neutron_lbaas.drivers import driver_mixins
from neutron_lbaas.services.loadbalancer import constants


class NotImplementedManager(object):
    """Helper class to make any subclass of LoadBalancerBaseDriver explode if
    it is missing any of the required object managers.
    """

    def create(self, context, obj):
        raise NotImplementedError()

    def update(self, context, old_obj, obj):
        raise NotImplementedError()

    def delete(self, context, obj):
        raise NotImplementedError()


class LoadBalancerBaseDriver(object):
    """LBaaSv2 object model drivers should subclass LoadBalancerBaseDriver,
    and initialize the following manager classes to create, update, and delete
    the various load balancer objects.
    """

    model_map = {constants.LOADBALANCER_EVENT: models.LoadBalancer,
                 constants.LISTENER_EVENT: models.Listener,
                 constants.POOL_EVENT: models.PoolV2,
                 constants.MEMBER_EVENT: models.MemberV2}

    load_balancer = NotImplementedManager()
    listener = NotImplementedManager()
    pool = NotImplementedManager()
    member = NotImplementedManager()
    health_monitor = NotImplementedManager()
    l7policy = NotImplementedManager()
    l7rule = NotImplementedManager()

    def __init__(self, plugin):
        self.plugin = plugin

    def handle_streamed_event(self, container):
        if container.info_type not in LoadBalancerBaseDriver.model_map:
            if container.info_type == constants.LOADBALANCER_STATS_EVENT:
                context = ncontext.get_admin_context()
                self.plugin.db.update_loadbalancer_stats(
                    context, container.info_id, container.info_payload)
            elif container.info_type == constants.LISTENER_STATS_EVENT:
                return
            else:
                exc = exceptions.ModelMapException(
                    target_name=container.info_type)
                raise exc
        else:
            model_class = LoadBalancerBaseDriver.model_map[
                container.info_type]
            context = ncontext.get_admin_context()
            self.plugin.db.update_status(context, model_class,
                                         container.info_id,
                                         **container.info_payload)


class BaseLoadBalancerManager(driver_mixins.BaseRefreshMixin,
                              driver_mixins.BaseStatsMixin,
                              driver_mixins.BaseManagerMixin):
    model_class = models.LoadBalancer

    @property
    def allows_create_graph(self):
        """
        Can this driver create a load balancer graph in one call.

        Return True if this driver has the capability to create a load balancer
        and any of its children in one driver call.  If this returns True and
        the user requests the creation of a load balancer graph, then the
        create_graph method will be called to create the load balancer.
        """
        return False

    @property
    def allows_healthmonitor_thresholds(self):
        """Does this driver support thresholds for health monitors"""
        return False

    @property
    def allocates_vip(self):
        """Does this driver need to allocate its own virtual IPs"""
        return False

    def create_and_allocate_vip(self, context, obj):
        """Create the load balancer and allocate a VIP

        If this method is implemented AND allocates_vip returns True, then
        this method will be called instead of the create method.  Any driver
        that implements this method is responsible for allocating a virtual IP
        and updating at least the vip_address attribute in the loadbalancer
        database table.
        """
        raise NotImplementedError

    @property
    def db_delete_method(self):
        return self.driver.plugin.db.delete_loadbalancer


class BaseListenerManager(driver_mixins.BaseManagerMixin):
    model_class = models.Listener

    @property
    def db_delete_method(self):
        return self.driver.plugin.db.delete_listener


class BasePoolManager(driver_mixins.BaseManagerMixin):
    model_class = models.PoolV2

    @property
    def db_delete_method(self):
        return self.driver.plugin.db.delete_pool


class BaseMemberManager(driver_mixins.BaseManagerMixin):
    model_class = models.MemberV2

    @property
    def db_delete_method(self):
        return self.driver.plugin.db.delete_member


class BaseHealthMonitorManager(driver_mixins.BaseManagerMixin):
    model_class = models.HealthMonitorV2

    @property
    def db_delete_method(self):
        return self.driver.plugin.db.delete_healthmonitor


class BaseL7PolicyManager(driver_mixins.BaseManagerMixin):
    model_class = models.L7Policy

    @property
    def db_delete_method(self):
        return self.driver.plugin.db.delete_l7policy


class BaseL7RuleManager(driver_mixins.BaseManagerMixin):
    model_class = models.L7Rule

    @property
    def db_delete_method(self):
        return self.driver.plugin.db.delete_l7policy_rule


# A decorator for wrapping driver operations, which will automatically
# set the neutron object's status based on whether it sees an exception

def driver_op(func):
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        d = (func.__name__ == 'delete')
        lb_create = ((func.__name__ == 'create') and
                     isinstance(args[0], BaseLoadBalancerManager))
        try:
            r = func(*args, **kwargs)
            args[0].successful_completion(
                args[1], args[2], delete=d, lb_create=lb_create)
            return r
        except Exception:
            with excutils.save_and_reraise_exception():
                args[0].failed_completion(args[1], args[2])
    return func_wrapper



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""edge_driver

Revision ID: 4ba00375f715
Revises: lbaasv2_tls
Create Date: 2015-02-03 20:35:54.830634

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '4ba00375f715'
down_revision = 'lbaasv2_tls'


def upgrade():
    op.create_table(
        'nsxv_edge_pool_mappings',
        sa.Column('pool_id', sa.String(length=36), nullable=False),
        sa.Column('edge_id', sa.String(length=36), nullable=False),
        sa.Column('edge_pool_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['pool_id'], ['pools.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('pool_id')
    )
    op.create_table(
        'nsxv_edge_vip_mappings',
        sa.Column('pool_id', sa.String(length=36), nullable=False),
        sa.Column('edge_id', sa.String(length=36), nullable=False),
        sa.Column('edge_app_profile_id', sa.String(length=36),
                  nullable=False),
        sa.Column('edge_vse_id', sa.String(length=36), nullable=False),
        sa.Column('edge_fw_rule_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['pool_id'], ['pools.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('pool_id')
    )
    op.create_table(
        'nsxv_edge_monitor_mappings',
        sa.Column('monitor_id', sa.String(length=36), nullable=False),
        sa.Column('edge_id', sa.String(length=36), nullable=False),
        sa.Column('edge_monitor_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['monitor_id'], ['healthmonitors.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('monitor_id'),
        sa.UniqueConstraint('monitor_id', 'edge_id',
                            name='uniq_nsxv_edge_monitor_mappings')
    )



# Copyright 2013 OpenStack Foundation.  All rights reserved
# Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class AgentDeviceDriver(object):
    """Abstract device driver that defines the API required by LBaaS agent."""

    def __init__(self, conf, plugin_rpc, process_monitor=None):
        self.conf = conf
        self.plugin_rpc = plugin_rpc
        self.process_monitor = process_monitor

    @abc.abstractproperty
    def loadbalancer(self):
        pass

    @abc.abstractproperty
    def listener(self):
        pass

    @abc.abstractproperty
    def pool(self):
        pass

    @abc.abstractproperty
    def member(self):
        pass

    @abc.abstractproperty
    def healthmonitor(self):
        pass

    @abc.abstractmethod
    def get_name(self):
        """Returns unique name across all LBaaS device drivers."""
        pass

    @abc.abstractmethod
    def deploy_instance(self, loadbalancer):
        """Fully deploys a loadbalancer instance from a given loadbalancer."""
        pass

    @abc.abstractmethod
    def undeploy_instance(self, loadbalancer_id, **kwargs):
        """Fully undeploys the loadbalancer instance."""
        pass

    def remove_orphans(self, known_loadbalancer_ids):
        # Not all drivers will support this
        raise NotImplementedError()


@six.add_metaclass(abc.ABCMeta)
class BaseManager(object):

    def __init__(self, driver):
        self.driver = driver

    @abc.abstractmethod
    def create(self, obj):
        pass

    @abc.abstractmethod
    def update(self, old_obj, obj):
        pass

    @abc.abstractmethod
    def delete(self, obj):
        pass


class BaseLoadBalancerManager(BaseManager):

    @abc.abstractmethod
    def get_stats(self, loadbalancer_id):
        pass


class BaseListenerManager(BaseManager):
    pass


class BasePoolManager(BaseManager):
    pass


class BaseMemberManager(BaseManager):
    pass


class BaseHealthMonitorManager(BaseManager):
    pass



# Copyright 2016 Rackspace Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils

from neutron_lbaas.tests.tempest.v2.scenario import base


class TestSharedPools(base.BaseTestCase):

    @utils.services('compute', 'network')
    def test_shared_pools(self):
        """This test checks load balancing with shared pools.

        The following is the scenario outline:
        1. Boot 1 instance.
        2. SSH to instance and start two servers listening on different ports.
        3. Create a pool.
        4. Create 2 listeners.
        5. Create a load balancer with two members associated to the two
           servers on the instance, and with ROUND_ROBIN algorithm.
        6. Send NUM requests to one listener's floating ip and check that
           they are shared between the two members.
        7. Send NUM requests to the other listener's floating ip and check that
           they are shared between the two members.
        """
        second_listener_port = 8080

        self._create_server('server1')
        self._start_servers()
        # automatically creates first listener on port 80
        self._create_load_balancer()
        # create second listener
        self._create_listener(load_balancer_id=self.load_balancer['id'],
                              port=second_listener_port,
                              default_pool_id=self.pool['id'])
        self._create_security_group_rules_for_port(second_listener_port)
        self._wait_for_load_balancer_status(self.load_balancer['id'])
        # check via first listener's default port 80
        self._check_load_balancing()
        # check via second listener's port
        self._check_load_balancing(port=second_listener_port)



# Copyright 2014, 2015 Rackspace US, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from barbicanclient import client as barbican_client
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from stevedore import driver as stevedore_driver

from neutron_lbaas._i18n import _
from neutron_lbaas.common.cert_manager import cert_manager

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class Cert(cert_manager.Cert):
    """Representation of a Cert based on the Barbican CertificateContainer."""

    def __init__(self, cert_container):
        if not isinstance(cert_container,
                          barbican_client.containers.CertificateContainer):
            raise TypeError(_(
                "Retrieved Barbican Container is not of the correct type "
                "(certificate)."))
        self._cert_container = cert_container

    # Container secrets are accessed upon query and can return as None,
    # don't return the payload if the secret is not available.

    def get_certificate(self):
        if self._cert_container.certificate:
            return self._cert_container.certificate.payload

    def get_intermediates(self):
        if self._cert_container.intermediates:
            return self._cert_container.intermediates.payload

    def get_private_key(self):
        if self._cert_container.private_key:
            return self._cert_container.private_key.payload

    def get_private_key_passphrase(self):
        if self._cert_container.private_key_passphrase:
            return self._cert_container.private_key_passphrase.payload


class CertManager(cert_manager.CertManager):
    """Certificate Manager that wraps the Barbican client API."""

    def __init__(self):
        super(CertManager, self).__init__()
        self.auth = stevedore_driver.DriverManager(
            namespace='neutron_lbaas.cert_manager.barbican_auth',
            name=cfg.CONF.certificates.barbican_auth,
            invoke_on_load=True,
        ).driver

    def store_cert(self, project_id, certificate, private_key,
                   intermediates=None, private_key_passphrase=None,
                   expiration=None, name='LBaaS TLS Cert'):
        """Stores a certificate in the certificate manager.

        :param certificate: PEM encoded TLS certificate
        :param private_key: private key for the supplied certificate
        :param intermediates: ordered and concatenated intermediate certs
        :param private_key_passphrase: optional passphrase for the supplied key
        :param expiration: the expiration time of the cert in ISO 8601 format
        :param name: a friendly name for the cert

        :returns: the container_ref of the stored cert
        :raises Exception: if certificate storage fails
        """

        connection = self.auth.get_barbican_client(project_id)

        LOG.info((
            "Storing certificate container '{0}' in Barbican."
        ).format(name))

        certificate_secret = None
        private_key_secret = None
        intermediates_secret = None
        pkp_secret = None

        try:
            certificate_secret = connection.secrets.create(
                payload=certificate,
                expiration=expiration,
                name="Certificate"
            )
            private_key_secret = connection.secrets.create(
                payload=private_key,
                expiration=expiration,
                name="Private Key"
            )
            certificate_container = connection.containers.create_certificate(
                name=name,
                certificate=certificate_secret,
                private_key=private_key_secret
            )
            if intermediates:
                intermediates_secret = connection.secrets.create(
                    payload=intermediates,
                    expiration=expiration,
                    name="Intermediates"
                )
                certificate_container.intermediates = intermediates_secret
            if private_key_passphrase:
                pkp_secret = connection.secrets.create(
                    payload=private_key_passphrase,
                    expiration=expiration,
                    name="Private Key Passphrase"
                )
                certificate_container.private_key_passphrase = pkp_secret

            certificate_container.store()
            return certificate_container.container_ref
        # Barbican (because of Keystone-middleware) sometimes masks
        #  exceptions strangely -- this will catch anything that it raises and
        #  reraise the original exception, while also providing useful
        #  feedback in the logs for debugging
        except Exception:
            for secret in [certificate_secret, private_key_secret,
                           intermediates_secret, pkp_secret]:
                if secret and secret.secret_ref:
                    old_ref = secret.secret_ref
                    try:
                        secret.delete()
                        LOG.info((
                            "Deleted secret {0} ({1}) during rollback."
                        ).format(secret.name, old_ref))
                    except Exception:
                        LOG.warning((
                            "Failed to delete {0} ({1}) during rollback. This "
                            "is probably not a problem."
                        ).format(secret.name, old_ref))
            with excutils.save_and_reraise_exception():
                LOG.exception("Error storing certificate data")

    def get_cert(self, project_id, cert_ref, resource_ref,
                 check_only=False, service_name='lbaas'):
        """Retrieves the specified cert and registers as a consumer.

        :param cert_ref: the UUID of the cert to retrieve
        :param resource_ref: Full HATEOAS reference to the consuming resource
        :param check_only: Read Certificate data without registering
        :param service_name: Friendly name for the consuming service

        :returns: octavia.certificates.common.Cert representation of the
                 certificate data
        :raises Exception: if certificate retrieval fails
        """
        connection = self.auth.get_barbican_client(project_id)

        LOG.info((
            "Loading certificate container {0} from Barbican."
        ).format(cert_ref))
        try:
            if check_only:
                cert_container = connection.containers.get(
                    container_ref=cert_ref
                )
            else:
                cert_container = connection.containers.register_consumer(
                    container_ref=cert_ref,
                    name=service_name,
                    url=resource_ref
                )
            return Cert(cert_container)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Error getting {0}".format(cert_ref))

    def delete_cert(self, project_id, cert_ref, resource_ref,
                    service_name='lbaas'):
        """Deregister as a consumer for the specified cert.

        :param cert_ref: the UUID of the cert to retrieve
        :param service_name: Friendly name for the consuming service
        :param lb_id: Loadbalancer id for building resource consumer URL

        :raises Exception: if deregistration fails
        """
        connection = self.auth.get_barbican_client(project_id)

        LOG.info((
            "Deregistering as a consumer of {0} in Barbican."
        ).format(cert_ref))
        try:
            connection.containers.remove_consumer(
                container_ref=cert_ref,
                name=service_name,
                url=resource_ref
            )
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception((
                    "Error deregistering as a consumer of {0}"
                ).format(cert_ref))
